# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations
from anyblok.column import String, Integer
from ..exceptions import CacheException


register = Declarations.register
System = Declarations.Model.System


@register(System)
class Cache:

    last_cache_id = None
    lrus = {}

    id = Integer(primary_key=True)
    registry_name = String(nullable=False)
    method = String(nullable=False)

    @classmethod
    def get_last_id(cls):
        """ Return the last primary key ``id`` value
        """
        res = cls.query('id').order_by(cls.id.desc()).limit(1).first()
        if res:
            return res[0]

        return 0

    @classmethod
    def initialize_model(cls):
        """ Initialize the last_cache_id known
        """
        super(Cache, cls).initialize_model()
        cls.last_cache_id = cls.get_last_id()

    @classmethod
    def invalidate_all(cls):
        res = []
        for registry_name, methods in cls.registry.caches.items():
            for method in methods.keys():
                res.append(dict(registry_name=registry_name, method=method))

        if res:
            cls.multi_insert(*res)

        cls.clear_invalidate_cache()

    @classmethod
    def invalidate(cls, registry_name, method):
        """ Call the invalidation for a specific method cached on a model

        :param registry_name: namespace of the model
        :param method: name of the method on the model
        :exception: CacheException
        """
        caches = cls.registry.caches

        def insert(registry_name=None, method=None):
            if registry_name in caches:
                if method in caches[registry_name]:
                    cls.insert(registry_name=registry_name, method=method)
                else:
                    raise CacheException(
                        "Unknown cached method %r" % method)
            else:
                raise CacheException(
                    "Unknown cached model %r" % registry_name)

        if isinstance(registry_name, str):
            insert(registry_name=registry_name, method=method)
        elif hasattr(registry_name, '__registry_name__'):
            insert(registry_name=registry_name.__registry_name__,
                   method=method)

        cls.clear_invalidate_cache()

    @classmethod
    def detect_invalidation(cls):
        """ Return True if a new invalidation is found in the table

        :rtype: Boolean
        """
        return cls.last_cache_id < cls.get_last_id()

    @classmethod
    def get_invalidation(cls):
        """ Return the pointer of the method to invalidate
        """
        res = []
        if cls.detect_invalidation():
            caches = cls.registry.caches
            for i in cls.query().filter(cls.id > cls.last_cache_id).all():
                res.extend(caches[i.registry_name][i.method])

            cls.last_cache_id = cls.get_last_id()

        return res

    @classmethod
    def clear_invalidate_cache(cls):
        """ Invalidate the cache that needs to be invalidated
        """
        for cache in cls.get_invalidation():
            cache.cache_clear()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer
from ..authorization import TestRuleOne, TestRuleTwo

register = Declarations.register
Model = Declarations.Model


@register(Model)
class Test:

    test2 = Integer(foreign_key=Model.Test2.use('id'))


Declarations.AuthorizationBinding(Declarations.Model.Test,
                                  TestRuleOne(),
                                  permission='Other')
Declarations.AuthorizationBinding(Declarations.Model.Test,
                                  TestRuleOne())
Declarations.AuthorizationBinding(Declarations.Model.Test,
                                  TestRuleTwo(),
                                  permission='Write')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String
from sqlalchemy import ForeignKeyConstraint


register = Declarations.register
System = Declarations.Model.System
Mixin = Declarations.Mixin


@register(System)  # noqa
class Field:

    name = String(primary_key=True)
    code = String(nullable=True)
    model = String(primary_key=True)
    # FIXME, foreign_key=System.Model.use('name'))
    label = String()
    ftype = String(label="Type", nullable=True)
    entity_type = String(nullable=True)

    @classmethod
    def define_table_args(cls):
        table_args = super(Field, cls).define_table_args()
        if cls.__registry_name__ != System.Field.__registry_name__:
            F = cls.registry.System.Field
            return table_args + (ForeignKeyConstraint([cls.name, cls.model],
                                                      [F.name, F.model],
                                                      ondelete="CASCADE"),)

        return table_args

    @classmethod
    def define_mapper_args(cls):
        mapper_args = super(Field, cls).define_mapper_args()
        if cls.__registry_name__ == System.Field.__registry_name__:
            mapper_args.update({
                'polymorphic_identity': cls.__registry_name__,
                'polymorphic_on': cls.entity_type,
            })
        else:
            mapper_args.update({
                'polymorphic_identity': cls.__registry_name__,
            })
        return mapper_args

    @classmethod
    def get_cname(self, field, cname):
        return cname

    def _description(self):
        res = {
            'id': self.name,
            'label': self.label,
            'type': self.ftype,
            'nullable': True,
            'primary_key': False,
            'model': None,
        }
        c = self.registry.loaded_namespaces_first_step[self.model][self.name]
        c.update_description(self.registry, self.model, res)
        return res

    @classmethod
    def add_field(cls, rname, label, model, table, ftype):
        """ Insert a field definition

        :param rname: name of the field
        :param label: label of the field
        :param model: namespace of the model
        :param table: name of the table of the model
        :param ftype: type of the AnyBlok Field
        """
        cls.insert(code=table + '.' + rname, model=model, name=rname,
                   label=label, ftype=ftype)

    @classmethod
    def alter_field(cls, field, label, ftype):
        """ Update an existing field

        :param field: instance of the Field model to update
        :param label: label of the field
        :param ftype: type of the AnyBlok Field
        """
        field.update(label=label, ftype=ftype)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.field import FieldException
from anyblok.relationship import RelationShip


class OneRelationShip(RelationShip):
    pass


class OneModel:
    __tablename__ = 'test'
    __registry_name__ = 'One.Model'


class MockRegistry:

    InstrumentedList = []


class TestRelationShip(TestCase):

    def test_forbid_instance(self):
        try:
            RelationShip(model=OneModel)
            self.fail("RelationShip mustn't be instanciated")
        except FieldException:
            pass

    def test_must_have_a_model(self):
        OneRelationShip(model=OneModel)
        try:
            OneRelationShip()
            self.fail("No watchdog, the model must be required")
        except FieldException:
            pass



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Per model flat access rule, based on a Model (table)"""
from anyblok.registry import RegistryManagerException
from .base import AuthorizationRule


class ModelAccessRule(AuthorizationRule):
    """Rule to grant authorization uniformly for all records of one model.

    The grants are themselves stored using a model class, that's provided
    in this blok. The users don't need to install the blok to use this class,
    provided they pass the model class to be used in all cases.
    """

    grant_model_name = 'Model.Authorization.ModelPermissionGrant'

    def __init__(self, grant_model=None):
        """.

        :params: grant_model is a model declaration, that has the needed
                 columns (model, principal, permission)
        """
        if grant_model is not None:
            self.grant_model_name = grant_model.__registry_name__

    @property
    def grant_model(self):
        try:
            return self.registry.get(self.grant_model_name)
        except RegistryManagerException:
            cls = self.__class__
            if self.grant_model_name is not cls.grant_model_name:
                raise
            raise RuntimeError(
                "To use %s with no explicit Grant "
                "model, you must install the model_access blok, "
                "that provides the default %r" % (
                    cls.__name__,
                    cls.grant_model_name))

    def check_on_model(self, model, principals, permission):
        Grant = self.grant_model
        return bool(Grant.query().filter(
            Grant.model == model,
            Grant.principal.in_(principals),
            Grant.permission == permission).limit(1).count())

    def check(self, record, principals, permission):
        return self.check_on_model(record.__registry_name__,
                                   principals,
                                   permission)

    def filter(self, model, query, principals, permission):
        if self.check_on_model(model.__registry_name__,
                               principals, permission):
            return query
        return False



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Selection
from csv import DictReader
from io import StringIO
from .exceptions import CSVImporterException


register = Declarations.register
Mixin = Declarations.Mixin
IO = Declarations.Model.IO


@register(IO)
class Importer(Mixin.IOCSVMixin):

    csv_on_error = Selection(
        selections=[('raise_now', 'Raise now'),
                    ('raise_at_the_end', 'Raise at the end'),
                    ('ignore', 'Ignore and continue')],
        default='raise_at_the_end')
    csv_if_exist = Selection(selections=[('pass', 'Pass to the next record'),
                                         ('overwrite', 'Update the record'),
                                         ('create', 'Create another record'),
                                         ('raise', 'Raise an exception')],
                             default='overwrite')
    csv_if_does_not_exist = Selection(selections=[
        ('pass', 'Pass to the next record'),
        ('create', 'Create the record'),
        ('raise', 'Raise an exception')], default='create')

    @classmethod
    def get_mode_choices(cls):
        res = super(Importer, cls).get_mode_choices()
        res.update({'Model.IO.Importer.CSV': 'CSV'})
        return res


@register(IO.Importer)
class CSV:

    def __init__(self, importer, blokname=None):
        self.importer = importer
        self.error_found = []
        self.reader = None
        self.created_entries = []
        self.updated_entries = []
        self.header_pks = []
        self.header_external_id = None
        self.header_external_ids = {}
        self.header_fields = []
        self.fields_description = {}
        self.blokname = blokname

    def commit(self):
        if self.error_found:
            return False

        self.importer.offset = self.reader.line_num - 1
        self.importer.commit()
        return True

    def get_reader(self):
        csvfile = StringIO()
        csvfile.write(self.importer.file_to_import.decode('utf-8'))
        csvfile.seek(0)
        self.reader = DictReader(csvfile,
                                 delimiter=self.importer.csv_delimiter,
                                 quotechar=self.importer.csv_quotechar)

    def consume_offset(self):
        try:
            for offset in range(self.importer.offset):
                next(self.reader)
        except StopIteration:
            pass

    def consume_nb_grouped_lines(self):
        res = []
        try:
            for offset in range(self.importer.nb_grouped_lines):
                res.append(next(self.reader))
        except StopIteration:
            pass

        return res

    def get_header(self):
        headers = self.reader.fieldnames
        Model = self.registry.get(self.importer.model)
        self.fields_description = Model.fields_description(
            fields=[h.split('/')[0] for h in headers])

        for header in headers:
            if '/' in header:
                name = header.split('/')[0]
                external_id = True
            else:
                name = header
                external_id = False

            if external_id:
                if not name or self.fields_description[name]['primary_key']:
                    self.header_external_id = header
                else:
                    self.header_external_ids[header] = name
            else:
                if self.fields_description[name]['primary_key']:
                    self.header_pks.append(header)
                else:
                    self.header_fields.append(name)

    def _parse_row_if_entry(self, row, entry, values, Model):
        if self.importer.csv_if_exist == 'overwrite':
            entry.update(**values)
            self.updated_entries.append(entry)
        elif self.importer.csv_if_exist == 'create':
            entry = Model.insert(**values)
            self.created_entries.append(entry)
        elif self.importer.csv_if_exist == 'raise':
            raise CSVImporterException(
                "Row %r already an entry %r " % (
                    row, entry.to_primary_keys()))

    def _parse_row_if_not_entry(self, row, pks, values, Model):
        Mapping = self.registry.IO.Mapping
        if self.importer.csv_if_does_not_exist == 'create':
            if pks:
                values.update(**pks)

            entry = Model.insert(**values)
            self.created_entries.append(entry)
            if self.header_external_id:
                Mapping.set(row[self.header_external_id], entry,
                            blokname=self.blokname)

        elif self.importer.csv_if_does_not_exist == 'raise':
            raise CSVImporterException(
                "Create row are not allowed")

    def _parse_row(self, row, entry, pks, values, Model):
        if entry:
            self._parse_row_if_entry(row, entry, values, Model)
        else:
            self._parse_row_if_not_entry(row, pks, values, Model)

    def parse_row(self, row):
        try:
            entry = pks = None
            Model = self.registry.get(self.importer.model)
            values = {}
            for field in self.header_fields:
                ctype = self.fields_description[field]['type']
                values[field] = self.importer.str2value(row[field], ctype)

            for external_field, field in self.header_external_ids.items():
                ctype = self.fields_description[field]['type']
                model = self.fields_description[field]['model']
                values[field] = self.importer.str2value(
                    row[external_field], ctype, external_id=True, model=model)

            if self.header_external_id:
                entry = self.importer.get_key_mapping(
                    row[self.header_external_id])
            elif self.header_pks:
                pks = {}
                for field in self.header_pks:
                    ctype = self.fields_description[field]['type']
                    pks[field] = self.importer.str2value(row[field], ctype)

                entry = Model.from_primary_keys(**pks)

            self._parse_row(row, entry, pks, values, Model)
        except Exception as e:
            msg = '%r: %r' % (e.__class__.__name__, e)
            self.error_found.append(msg)
            if self.importer.csv_on_error == 'raise_now':
                raise CSVImporterException(msg)

    def run(self):
        try:
            self.get_reader()
            self.get_header()
            self.consume_offset()
            while True:
                rows = self.consume_nb_grouped_lines()
                if not rows:
                    break

                for row in rows:
                    self.parse_row(row)

                self.commit()
        except Exception as e:
            msg = '%r: %r' % (e.__class__.__name__, e)
            self.error_found.append(msg)

        if self.error_found:
            if self.importer.csv_on_error == 'raise_at_the_end':
                msg = "Exception found : \n %s" % '\n'.join(self.error_found)
                raise CSVImporterException(msg)

        return {
            'error': self.error_found,
            'created_entries': self.created_entries,
            'updated_entries': self.updated_entries,
        }

    @classmethod
    def insert(cls, delimiter=None, quotechar=None, **kwargs):
        kwargs['mode'] = cls.__registry_name__
        if 'model' not in kwargs:
            raise CSVImporterException("The column 'model' is required")

        if not isinstance(kwargs['model'], str):
            kwargs['model'] = kwargs['model'].__registry_name__

        if delimiter is not None:
            kwargs['csv_delimiter'] = delimiter

        if quotechar is not None:
            kwargs['csv_quotechar'] = quotechar

        return cls.registry.IO.Importer.insert(**kwargs)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations, reload_module_if_blok_is_reloading
from anyblok.config import Configuration


@Declarations.register(Declarations.Model.Documentation)
class Model(Declarations.Mixin.DocElement):

    def __init__(self, model):
        self.model = model
        self.fields = []
        self.attributes = []
        if self.exist():
            self._auto_doc(
                self.registry.Documentation.Model.Field, self.fields, self)
            self._auto_doc(self.registry.Documentation.Model.Attribute,
                           self.attributes, self)

    def exist(self):
        return self.registry.has(self.model.name)

    @classmethod
    def filterModel(cls, query):
        Model = cls.registry.System.Model
        wanted_models = Configuration.get('doc_wanted_models')
        if wanted_models:
            query = query.filter(Model.name.in_(wanted_models))
        unwanted_models = Configuration.get('doc_unwanted_models')
        if unwanted_models:
            query = query.filter(Model.name.notin_(unwanted_models))

        return query

    @classmethod
    def getelements(cls):
        return cls.filterModel(cls.registry.System.Model.query()).all()

    @classmethod
    def header2RST(cls, doc):
        doc.write("Models\n======\n\n"
                  "This the differents models defined "
                  "on the project" + ('\n' * 2))

    @classmethod
    def footer2RST(cls, doc):
        pass

    def toRST(self, doc):
        doc.write(self.model.name + '\n' + '-' * len(self.model.name) + '\n\n')
        self.toRST_docstring(doc)
        self.toRST_properties(doc)
        self.toRST_field(doc)
        self.toRST_method(doc)

    def toRST_field(self, doc):
        if self.fields:
            self._toRST(
                doc, self.registry.Documentation.Model.Field, self.fields)

    def toRST_method(self, doc):
        if self.attributes:
            self._toRST(
                doc, self.registry.Documentation.Model.Attribute,
                self.attributes)

    def toRST_docstring(self, doc):
        Model = self.registry.get(self.model.name)
        if hasattr(Model, '__doc__') and Model.__doc__:
            doc.write(Model.__doc__ + '\n\n')

    def toRST_properties_get(self):
        Model = self.registry.get(self.model.name)
        tablename = getattr(Model, '__tablename__', 'No table')
        return {
            'table name': tablename,
        }

    def toRST_properties(self, doc):
        properties = self.toRST_properties_get()
        msg = 'Properties:\n\n* ' + '\n* '.join('**%s** : %s' % (x, y)
                                                for x, y in properties.items())
        doc.write(msg + '\n\n')

    def toUML_add_model(self, dot):
        dot.add_class(self.model.name)

    def toUML_add_attributes(self, dot):
        for f in self.fields:
            f.toUML(dot)

        for attr in self.attributes:
            attr.toUML(dot, self.model.name)

    def toSQL_add_table(self, dot):
        Model = self.registry.get(self.model.name)
        if hasattr(Model, '__tablename__'):
            dot.add_table(Model.__tablename__)

    def toSQL_add_fields(self, dot):
        Model = self.registry.get(self.model.name)
        if hasattr(Model, '__tablename__'):
            for f in self.fields:
                f.toSQL(dot)


from . import field  # noqa
reload_module_if_blok_is_reloading(field)
from . import attribute  # noqa
reload_module_if_blok_is_reloading(attribute)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .mapper import MapperAdapter
from .common import add_autodocs


class DeclarationsException(AttributeError):
    """ Simple Exception for Declarations """


class Declarations:
    """ Represents all the declarations done by the bloks

    .. warning::
        This is a global information, during the execution you must use the
        registry. The registry is the real assembler of the python classes
        based on the installed bloks

    ::

        from anyblok import Declarations

    """
    declaration_types = {}

    @classmethod
    def register(cls, parent, cls_=None, **kwargs):
        """ Method to add the blok in the registry under a type of declaration

        :param parent: An existing blok class in the Declaration
        :param ``cls_``: The ``class`` object to add in the Declaration
        :rtype: ``cls_``
        :exception: DeclarationsException
        """

        def wrapper(self):
            name = kwargs.get('name_', self.__name__)
            if parent.__declaration_type__ not in cls.declaration_types:
                raise DeclarationsException(
                    "No parents %r for %s" % (parent, name))

            declaration = cls.declaration_types[parent.__declaration_type__]
            declaration.register(parent, name, self, **kwargs)

            node = getattr(parent, name)
            setattr(node, '__declaration_type__', parent.__declaration_type__)
            setattr(node, '__registry_name__',
                    parent.__registry_name__ + '.' + name)

            # Only for auto doc with autoanyblok-declaration directive
            setattr(self, '__declaration__', declaration)
            setattr(self, '__registry_name__',
                    parent.__registry_name__ + '.' + name)
            return self

        if cls_:
            return wrapper(cls_)
        else:
            return wrapper

    @classmethod
    def unregister(cls, entry, cls_):
        """ Method to remove the blok from a type of declaration

        :param entry: declaration entry of the model where the ``cls_``
            must be removed
        :param ``cls_``: The ``class`` object to remove from the
            Declaration
        :rtype: ``cls_``
        """
        declaration = cls.declaration_types[entry.__declaration_type__]
        declaration.unregister(entry, cls_)

        return cls_

    @classmethod
    def add_declaration_type(cls, cls_=None, isAnEntry=False,
                             assemble=None, initialize=None, unload=None):
        """ Add a declaration type

        :param cls_: The ``class`` object to add as a world of the MetaData
        :param isAnEntry: if true the type will be assembled by the registry
        :param assemble: name of the method callback to call (classmethod)
        :param initialize: name of the method callback to call (classmethod)
        :param unload: name of the method callback to call (classmethod)
        :exception: DeclarationsException
        """

        def wrapper(self):
            from anyblok.registry import RegistryManager
            name = self.__name__
            if name in cls.declaration_types:
                raise DeclarationsException(
                    "The declaration type %r is already defined" % name)

            cls.declaration_types[name] = self

            setattr(self, '__registry_name__', name)
            setattr(self, '__declaration_type__', name)
            setattr(cls, name, self)

            if isAnEntry:
                assemble_callback = initialize_callback = None
                if assemble and hasattr(self, assemble):
                    assemble_callback = getattr(self, assemble)

                if initialize and hasattr(self, initialize):
                    initialize_callback = getattr(self, initialize)

                RegistryManager.declare_entry(
                    name, assemble_callback=assemble_callback,
                    initialize_callback=initialize_callback)

            # All declaration type can need to be unload declarated values
            if unload and hasattr(self, unload):
                RegistryManager.declare_unload_callback(
                    name, getattr(self, unload))

            return self

        if cls_:
            return wrapper(cls_)
        else:
            return wrapper


def cache(size=128):
    autodoc = """
    **Cached method** with size=%(size)s
    """ % dict(size=size)

    def wrapper(method):
        add_autodocs(method, autodoc)
        method.is_cache_method = True
        method.is_cache_classmethod = False
        method.size = size
        return method

    return wrapper


def classmethod_cache(size=128):
    autodoc = """
    **Cached classmethod** with size=%(size)s
    """ % dict(size=size)

    def wrapper(method):
        add_autodocs(method, autodoc)
        method.is_cache_method = True
        method.is_cache_classmethod = True
        method.size = size
        return method

    return wrapper


def hybrid_method(method=None):
    autodoc = """
    **Hybrid method**
    """

    if method:
        add_autodocs(method, autodoc)
        method.is_an_hybrid_method = True
        return method
    else:

        def wrapper(method):
            add_autodocs(method, autodoc)
            method.is_an_hybrid_method = True
            return method

        return wrapper


def listen(*args, **kwargs):
    autodoc = """
    **listen** event call with the arguments %(args)r and the positionnal
    argument %(kwargs)r
    """ % dict(args=args, kwargs=kwargs)

    mapper = MapperAdapter(*args, **kwargs)

    def wrapper(method):
        add_autodocs(method, autodoc)
        mapper.listen(method)
        return classmethod(method)

    return wrapper



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.registry import RegistryManager
from anyblok import Declarations
from anyblok.field import Field
from anyblok.relationship import RelationShip
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import table
from sqlalchemy.orm import Query, mapper, synonym
from sqlalchemy import inspection
from sqlalchemy.ext.hybrid import hybrid_method
from anyblok.common import TypeList, apply_cache
from copy import deepcopy
from sqlalchemy.ext.declarative import declared_attr
from .mapper import ModelAttribute
from sqlalchemy import ForeignKeyConstraint
from anyblok.common import anyblok_column_prefix
from texttable import Texttable


class ModelException(Exception):
    """Exception for Model declaration"""


class ViewException(ModelException):
    """Exception for View declaration"""


class CreateView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    def __init__(self, name):
        self.name = name


@compiles(CreateView)
def compile_create_view(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (
        element.name, compiler.sql_compiler.process(element.selectable))


@compiles(DropView)
def compile_drop_view(element, compiler, **kw):
    return "DROP VIEW IF EXISTS %s" % (element.name)


def has_sql_fields(bases):
    """ Tells whether the model as field or not

    :param bases: list of Model's Class
    :rtype: boolean
    """
    for base in bases:
        for p in base.__dict__.keys():
            if hasattr(getattr(base, p), '__class__'):
                if Field in getattr(base, p).__class__.__mro__:
                    return True

    return False


def has_sqlalchemy_fields(base):
    for p in base.__dict__.keys():
        attr = base.__dict__[p]
        if inspection.inspect(attr, raiseerr=False) is not None:
            return True

    return False


def get_fields(base, without_relationship=False, only_relationship=False):
    """ Return the fields for a model

    :param base: Model Class
    :param without_relationship: Do not return the relationship field
    :param only_relationship: return only the relationship field
    :rtype: dict with name of the field in key and instance of Field in value
    """
    fields = {}
    for p in base.__dict__.keys():
        if hasattr(getattr(base, p), '__class__'):
            if without_relationship:
                if RelationShip in getattr(base, p).__class__.__mro__:
                    continue

            if only_relationship:
                if RelationShip not in getattr(base, p).__class__.__mro__:
                    continue

            if Field in getattr(base, p).__class__.__mro__:
                fields[p] = getattr(base, p)

    return fields


@Declarations.add_declaration_type(isAnEntry=True,
                                   assemble='assemble_callback',
                                   initialize='initialize_callback')
class Model:
    """ The Model class is used to define or inherit an SQL table.

    Add new model class::

        @Declarations.register(Declarations.Model)
        class MyModelclass:
            pass

    Remove a model class::

        Declarations.unregister(Declarations.Model.MyModelclass,
                                MyModelclass)

    There are three Model families:

    * No SQL Model: These models have got any field, so any table
    * SQL Model:
    * SQL View Model: it is a model mapped with a SQL View, the insert, update
      delete method are forbidden by the database

    Each model has a:

    * registry name: compose by the parent + . + class model name
    * table name: compose by the parent + '_' + class model name

    The table name can be overloaded by the attribute tablename. the wanted
    value are a string (name of the table) of a model in the declaration.

    ..warning::

        Two models can have the same table name, both models are mapped on
        the table. But they must have the same column.
    """

    @classmethod
    def register(self, parent, name, cls_, **kwargs):
        """ add new sub registry in the registry

        :param parent: Existing global registry
        :param name: Name of the new registry to add it
        :param cls_: Class Interface to add in registry
        """
        _registryname = parent.__registry_name__ + '.' + name
        if 'tablename' in kwargs:
            tablename = kwargs.pop('tablename')
            if not isinstance(tablename, str):
                tablename = tablename.__tablename__

        elif hasattr(parent, name):
            tablename = getattr(parent, name).__tablename__
        else:
            if parent is Declarations or parent is Declarations.Model:
                tablename = name.lower()
            elif hasattr(parent, '__tablename__'):
                tablename = parent.__tablename__
                tablename += '_' + name.lower()

        if not hasattr(parent, name):
            p = {
                '__tablename__': tablename,
                '__registry_name__': _registryname,
                'use': lambda x: ModelAttribute(_registryname, x),
            }
            ns = type(name, tuple(), p)
            setattr(parent, name, ns)

        if parent is Declarations:
            return

        kwargs['__registry_name__'] = _registryname
        kwargs['__tablename__'] = tablename

        RegistryManager.add_entry_in_register(
            'Model', _registryname, cls_, **kwargs)
        setattr(cls_, '__anyblok_kwargs__', kwargs)

    @classmethod
    def unregister(self, entry, cls_):
        """ Remove the Interface from the registry

        :param entry: entry declaration of the model where the ``cls_``
            must be removed
        :param cls_: Class Interface to remove in registry
        """
        RegistryManager.remove_in_register(cls_)

    @classmethod
    def declare_field(self, registry, name, field, namespace, properties):
        """ Declare the field/column/relationship to put in the properties
        of the model

        :param registry: the current  registry
        :param name: name of the field / column or relationship
        :param field: the declaration field / column or relationship
        :param namespace: the namespace of the model
        :param properties: the properties of the model
        """
        if name in properties['loaded_columns']:
            return

        if field.must_be_duplicate_before_added():
            field = deepcopy(field)

        attr_name = name
        if field.use_hybrid_property:
            attr_name = anyblok_column_prefix + name

        if field.must_be_declared_as_attr():
            # All the declaration are seen as mixin for sqlalchemy
            # some of them need de be defered for the initialisation
            # cause of the mixin as relation ship and column with foreign key
            def wrapper(cls):
                return field.get_sqlalchemy_mapping(
                    registry, namespace, name, properties)

            properties[attr_name] = declared_attr(wrapper)
            properties[attr_name].anyblok_field = field
        else:
            properties[attr_name] = field.get_sqlalchemy_mapping(
                registry, namespace, name, properties)

        if field.use_hybrid_property:
            properties[name] = field.get_property(
                registry, namespace, name, properties)
            properties['hybrid_property_columns'].append(name)

        properties['loaded_columns'].append(name)
        field.update_properties(registry, namespace, name, properties)

    @classmethod
    def apply_event_listner(cls, attr, method, registry, namespace, base,
                            properties):
        """ Find the event listener methods in the base to save the
        namespace and the method in the registry

        :param attr: name of the attibute
        :param method: method pointer
        :param registry: the current  registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param properties: the properties of the model
        """
        if not hasattr(method, 'is_an_event_listener'):
            return
        elif method.is_an_event_listener is True:
            model = method.model
            event = method.event
            events = registry.events
            if model not in events:
                events[model] = {event: []}
            elif event not in events[model]:
                events[model][event] = []

            val = (namespace, attr)
            ev = events[model][event]
            if val not in ev:
                ev.append(val)

    @classmethod
    def apply_sqlalchemy_event_listner(cls, attr, method, registry, namespace,
                                       base, properties):
        """declare in the registry the sqlalchemy event

        :param attr: name of the attibute
        :param method: method pointer
        :param registry: the current  registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param properties: the properties of the model
        """
        if not hasattr(method, 'is_an_sqlalchemy_event_listener'):
            return
        elif method.is_an_sqlalchemy_event_listener is True:
            registry._sqlalchemy_known_events.append(
                (method.sqlalchemy_listener,
                 namespace,
                 ModelAttribute(namespace, attr)))

    @classmethod
    def detect_hybrid_method(cls, attr, method, registry, namespace, base,
                             properties):
        """ Find the sqlalchemy hybrid methods in the base to save the
        namespace and the method in the registry

        :param attr: name of the attibute
        :param method: method pointer
        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param properties: the properties of the model
        """
        if not hasattr(method, 'is_an_hybrid_method'):
            return
        elif method.is_an_hybrid_method is True:
            if attr not in properties['hybrid_method']:
                properties['hybrid_method'].append(attr)

    @classmethod
    def detect_table_and_mapper_args(cls, registry, namespace, base,
                                     properties):
        """Test if define_table/mapper_args are in the base, and call them
        save the value in the properties

        if  __table/mapper_args\_\_ are in the base then raise ModelException

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param properties: the properties of the model
        :exception: ModelException
        """
        if hasattr(base, '__table_args__'):
            raise ModelException(
                "'__table_args__' attribute is forbidden, on Model : %r (%r)."
                "Use the class method 'define_table_args' to define the value "
                "allow anyblok to fill his own '__table_args__' attribute" % (
                    namespace, base.__table_args__))

        if hasattr(base, '__mapper_args__'):
            raise ModelException(
                "'__mapper_args__' attribute is forbidden, on Model : %r (%r)."
                "Use the class method 'define_mapper_args' to define the "
                "value allow anyblok to fill his own '__mapper_args__' "
                "attribute" % (namespace, base.__mapper_args__))

        if hasattr(base, 'define_table_args'):
            properties['table_args'] = True

        if hasattr(base, 'define_mapper_args'):
            properties['mapper_args'] = True

    @classmethod
    def transform_base(cls, registry, namespace, base, properties):
        """ Detect specific declaration which must define by registry

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param properties: the properties of the model
        :rtype: new base
        """
        new_type_properties = {}
        for attr in dir(base):
            method = getattr(base, attr)
            new_type_properties.update(apply_cache(
                attr, method, registry, namespace, base, properties))
            cls.apply_event_listner(
                attr, method, registry, namespace, base, properties)
            cls.apply_sqlalchemy_event_listner(
                attr, method, registry, namespace, base, properties)
            cls.detect_hybrid_method(
                attr, method, registry, namespace, base, properties)

        cls.detect_table_and_mapper_args(
            registry, namespace, base, properties)

        if new_type_properties:
            return [type(namespace, (), new_type_properties), base]

        return [base]

    @classmethod
    def apply_hybrid_method(cls, base, registry, namespace, bases,
                            transformation_properties, properties):
        """ Create overload to define the write declaration of sqlalchemy
        hybrid method, add the overload in the declared bases of the
        namespace

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param properties: assembled attributes of the namespace
        """
        type_properties = {}

        def apply_wrapper(attr):

            def wrapper(self, *args, **kwargs):
                self_ = self.registry.loaded_namespaces[self.__registry_name__]
                if self is self_:
                    return getattr(super(base, self), attr)(
                        self, *args, **kwargs)
                else:
                    return getattr(super(base, self), attr)(
                        *args, **kwargs)

            setattr(base, attr, hybrid_method(wrapper))

        if transformation_properties['hybrid_method']:
            for attr in transformation_properties['hybrid_method']:
                apply_wrapper(attr)

        return type_properties

    @classmethod
    def apply_table_and_mapper_args(cls, base, registry, namespace, bases,
                                    transformation_properties, properties):
        """ Create overwrite to define table and mapper args to define some
        options for SQLAlchemy

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param properties: assembled attributes of the namespace
        """
        table_args = tuple(properties['add_in_table_args'])
        if table_args:
            def define_table_args(cls_):
                if cls_.__registry_name__ == namespace:
                    res = super(base, cls_).define_table_args()
                    fks = [x.name for x in res
                           if isinstance(x, ForeignKeyConstraint)]

                    t_args = [x for x in table_args
                              if (not isinstance(x, ForeignKeyConstraint) or
                                  x.name not in fks)]

                    return res + tuple(t_args)

                return ()

            base.define_table_args = classmethod(define_table_args)
            transformation_properties['table_args'] = True

        if transformation_properties['table_args']:

            def __table_args__(cls_):
                return cls_.define_table_args()

            base.__table_args__ = declared_attr(__table_args__)

        if transformation_properties['mapper_args']:

            def __mapper_args__(cls_):
                return cls_.define_mapper_args()

            base.__mapper_args__ = declared_attr(__mapper_args__)

    @classmethod
    def insert_in_bases(cls, registry, namespace, bases,
                        transformation_properties, properties):
        """ Add in the declared namespaces new base.

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param properties: assembled attributes of the namespace
        """
        new_base = type(namespace, (), {})
        bases.insert(0, new_base)
        cls.apply_hybrid_method(new_base, registry, namespace, bases,
                                transformation_properties, properties)
        cls.apply_table_and_mapper_args(new_base, registry, namespace, bases,
                                        transformation_properties, properties)

    @classmethod
    def raise_if_has_sqlalchemy(cls, base):
        if has_sqlalchemy_fields(base):
            raise ModelException(
                "the base %r have an SQLAlchemy attribute" % base)

    @classmethod
    def load_namespace_first_step(cls, registry, namespace):
        """ Return the properties of the declared bases for a namespace.
        This is the first step because some actions need to known all the
        properties

        :param registry: the current registry
        :param namespace: the namespace of the model
        :rtype: dict of the known properties
        """
        if namespace in registry.loaded_namespaces_first_step:
            return registry.loaded_namespaces_first_step[namespace]

        bases = []
        properties = {'__depends__': set()}
        ns = registry.loaded_registries[namespace]

        for b in ns['bases']:
            bases.append(b)

            for b_ns in b.__anyblok_bases__:
                if b_ns.__registry_name__.startswith('Model.'):
                    properties['__depends__'].add(b_ns.__registry_name__)

                ps = cls.load_namespace_first_step(registry,
                                                   b_ns.__registry_name__)
                ps.update(properties)
                properties.update(ps)

        for b in bases:
            cls.raise_if_has_sqlalchemy(b)
            fields = get_fields(b)
            for p, f in fields.items():
                if p not in properties:
                    properties[p] = f

        if '__tablename__' in ns['properties']:
            properties['__tablename__'] = ns['properties']['__tablename__']

        registry.loaded_namespaces_first_step[namespace] = properties
        return properties

    @classmethod
    def apply_inheritance_base(cls, registry, namespace, ns, bases,
                               realregistryname, properties,
                               transformation_properties):

        # remove doublon
        for b in ns['bases']:
            if b in bases:
                continue

            kwargs = {
                'namespace': realregistryname} if realregistryname else {}
            bases.append(b, **kwargs)

            if b.__doc__ and '__doc__' not in properties:
                properties['__doc__'] = b.__doc__

            for b_ns in b.__anyblok_bases__:
                brn = b_ns.__registry_name__
                if brn in registry.loaded_registries['Mixin_names']:
                    tp = transformation_properties
                    if realregistryname:
                        bs, ps = cls.load_namespace_second_step(
                            registry, brn, realregistryname=realregistryname,
                            transformation_properties=tp)
                    else:
                        bs, ps = cls.load_namespace_second_step(
                            registry, brn, realregistryname=namespace,
                            transformation_properties=tp)
                elif brn in registry.loaded_registries['Model_names']:
                    bs, ps = cls.load_namespace_second_step(registry, brn)
                else:
                    raise ModelException(
                        "You have not to inherit the %r "
                        "Only the 'Mixin' and %r types are allowed" % (
                            brn, cls.__name__))

                bases += bs

    @classmethod
    def init_core_properties_and_bases(cls, registry, bases, properties):
        properties['loaded_columns'] = []
        properties['hybrid_property_columns'] = []
        properties['loaded_fields'] = {}
        if properties['is_sql_view']:
            bases.extend([x for x in registry.loaded_cores['SqlViewBase']])
        elif has_sql_fields(bases):
            bases.extend([x for x in registry.loaded_cores['SqlBase']])
            bases.append(registry.declarativebase)
        else:
            # remove tablename to inherit from a sqlmodel
            del properties['__tablename__']

        bases.extend([x for x in registry.loaded_cores['Base']])

    @classmethod
    def declare_all_fields(cls, registry, namespace, bases, properties):
        # do in the first time the fields and columns
        # because for the relationship on the same model
        # the primary keys must exist before the relationship
        # load all the base before do relationship because primary key
        # can be come from inherit
        for b in bases:
            for p, f in get_fields(b,
                                   without_relationship=True).items():
                cls.declare_field(
                    registry, p, f, namespace, properties)

        for b in bases:
            for p, f in get_fields(b, only_relationship=True).items():
                cls.declare_field(
                    registry, p, f, namespace, properties)

    @classmethod
    def apply_existing_table(cls, registry, namespace, tablename, properties):
        if '__tablename__' in properties:
            del properties['__tablename__']

        for t in registry.loaded_namespaces.keys():
            m = registry.loaded_namespaces[t]
            if m.is_sql:
                if getattr(m, '__tablename__'):
                    if m.__tablename__ == tablename:
                        properties['__table__'] = m.__table__
                        tablename = namespace.replace('.', '_').lower()

    @classmethod
    def load_namespace_second_step(cls, registry, namespace,
                                   realregistryname=None,
                                   transformation_properties=None):
        """ Return the bases and the properties of the namespace

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param realregistryname: the name of the model if the namespace is a
            mixin
        :rtype: the list od the bases and the properties
        :exception: ModelException
        """
        if namespace in registry.loaded_namespaces:
            return [registry.loaded_namespaces[namespace]], {}

        if transformation_properties is None:
            transformation_properties = {
                'hybrid_method': [],
                'table_args': False,
                'mapper_args': False,
            }

        bases = TypeList(cls, registry, namespace, transformation_properties)
        ns = registry.loaded_registries[namespace]
        properties = ns['properties'].copy()
        properties['add_in_table_args'] = []

        if 'is_sql_view' not in properties:
            properties['is_sql_view'] = False

        cls.apply_inheritance_base(registry, namespace, ns, bases,
                                   realregistryname, properties,
                                   transformation_properties)

        if namespace in registry.loaded_registries['Model_names']:
            tablename = properties['__tablename__']
            cls.init_core_properties_and_bases(registry, bases, properties)

            if tablename in registry.declarativebase.metadata.tables:
                cls.apply_existing_table(
                    registry, namespace, tablename, properties)
            else:
                cls.declare_all_fields(registry, namespace, bases, properties)

            bases.append(registry.registry_base)
            cls.insert_in_bases(registry, namespace, bases,
                                transformation_properties, properties)
            if properties['is_sql_view']:
                cls.replace_properties_by_synonym(properties)
                bases = [type(tablename, tuple(bases), properties)]
                if properties['is_sql_view']:
                    cls.apply_view(namespace, tablename, bases[0], registry,
                                   properties)
            else:
                bases = [type(tablename, tuple(bases), properties)]

            properties = {}
            registry.add_in_registry(namespace, bases[0])
            registry.loaded_namespaces[namespace] = bases[0]

        return bases, properties

    @classmethod
    def replace_properties_by_synonym(cls, properties):
        for field in properties['loaded_columns']:
            properties[field] = synonym(anyblok_column_prefix + field)

    @classmethod
    def apply_view(cls, namespace, tablename, base, registry, properties):
        """ Transform the sqlmodel to view model

        :param namespace: Namespace of the model
        :param tablename: Name od the table of the model
        :param base: Model cls
        :param registry: current registry
        :param properties: properties of the model
        :exception: MigrationException
        :exception: ViewException
        """
        if hasattr(base, '__view__'):
            view = base.__view__
        elif tablename in registry.loaded_views:
            view = registry.loaded_views[tablename]
        else:
            if not hasattr(base, 'sqlalchemy_view_declaration'):
                raise ViewException(
                    "%r.'sqlalchemy_view_declaration' is required to "
                    "define the query to apply of the view" % namespace)

            view = table(tablename)
            registry.loaded_views[tablename] = view
            selectable = getattr(base, 'sqlalchemy_view_declaration')()

            if isinstance(selectable, Query):
                selectable = selectable.subquery()

            for c in selectable.c:
                c._make_proxy(view)

            DropView(tablename).execute_at(
                'before-create', registry.declarativebase.metadata)
            CreateView(tablename, selectable).execute_at(
                'after-create', registry.declarativebase.metadata)
            DropView(tablename).execute_at(
                'before-drop', registry.declarativebase.metadata)

        pks = [col for col in properties['loaded_columns']
               if getattr(base, anyblok_column_prefix + col).primary_key]

        if not pks:
            raise ViewException(
                "%r have any primary key defined" % namespace)

        pks = [getattr(view.c, x) for x in pks]
        mapper(base, view, primary_key=pks)
        setattr(base, '__view__', view)

    @classmethod
    def assemble_callback(cls, registry):
        """ Assemble callback is called to assemble all the Model
        from the installed bloks

        :param registry: registry to update
        """
        registry.loaded_namespaces_first_step = {}
        registry.loaded_views = {}
        registry.caches = {}
        registry.events = {}

        # get all the information to create a namespace
        for namespace in registry.loaded_registries['Model_names']:
            cls.load_namespace_first_step(registry, namespace)

        # create the namespace with all the information come from first
        # step
        for namespace in registry.loaded_registries['Model_names']:
            cls.load_namespace_second_step(registry, namespace)

    @classmethod
    def initialize_callback(cls, registry):
        """ initialize callback is called after assembling all entries

        This callback updates the database information about

        * Model
        * Column
        * RelationShip

        :param registry: registry to update
        """
        for Model in registry.loaded_namespaces.values():
            Model.initialize_model()

        Blok = registry.System.Blok
        if not registry.withoutautomigration:
            Model = registry.System.Model
            Model.update_list()
            registry.update_blok_list()

        bloks = Blok.list_by_state('touninstall')
        Blok.uninstall_all(*bloks)
        return Blok.apply_state(*registry.ordered_loaded_bloks)

    @classmethod
    def autodoc_class(cls, model_cls):
        res = [":Declaration type: Model"]
        res.extend([':%s: %s' % (x.replace('_', ' ').strip(), str(y))
                    for x, y in model_cls.__anyblok_kwargs__.items()])
        res.extend([':Inherit model or mixin:', ''])
        res.extend([' * ' + str(x) for x in model_cls.__anyblok_bases__])
        res.extend(['', ''])
        if has_sql_fields([model_cls]):
            rows = [['field name', 'Description']]
            rows.extend([x, y.autodoc()]
                        for x, y in get_fields(model_cls).items())
            table = Texttable()
            table.set_cols_valign(["m", "t"])
            table.add_rows(rows)
            res.extend(['', table.draw(), '', ''])

        return '\n'.join(res)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import anyblok
from anyblok.release import version
from anyblok.blok import BlokManager
from anyblok.config import Configuration
from anyblok.registry import RegistryManager
from anyblok._graphviz import ModelSchema, SQLSchema
from nose import main
import sys
from os.path import join, exists
from argparse import RawDescriptionHelpFormatter
from textwrap import dedent
from sqlalchemy_utils.functions import create_database
from anyblok import load_init_function_from_entry_points

Configuration.applications.update({
    'createdb': {
        'prog': 'AnyBlok create database, version %r' % version,
        'description': "Create a database and install bloks to populate it",
        'configuration_groups': ['config', 'database', 'unittest'],
    },
    'updatedb': {
        'prog': 'AnyBlok update database, version %r' % version,
        'description': ("Update a database: install, upgrade or uninstall the "
                        "bloks "),
        'configuration_groups': ['config', 'database', 'unittest'],
    },
    'nose': {
        'prog': 'AnyBlok nose, version %r' % version,
        'description': "Run fonctionnal nosetest of the installed bloks",
    },
    'interpreter': {
        'prog': 'AnyBlok interpretor, version %r' % version,
        'description': "Run an interpreter on the registry",
        'formatter_class': RawDescriptionHelpFormatter,
        'epilog': dedent("Example\n"
                         "-------\n"
                         "  $ anyblok_interpreter [anyblok arguments] \n"
                         "  $ => registry \n"
                         "  ... <registry> \n\n"
                         "  The interpretor add in the local the registry of "
                         "the selected database \n\n"
                         "Note\n"
                         "----\n"
                         "  if the ipython is in the python path, then "
                         "the interpretor will be an ipyton interpretor")
    },
    'worker': {
        'prog': 'AnyBlok cron worker, version %r' % version,
    },
})


def format_configuration(configuration_groups, *confs):
    if configuration_groups is None:
        configuration_groups = []

    for conf in confs:
        if conf not in configuration_groups:
            configuration_groups.append(conf)


def createdb(application, configuration_groups, **kwargs):
    """ Create a database and install blok from config

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    format_configuration(configuration_groups,
                         'create_db', 'install-bloks',
                         'install-or-update-bloks')
    load_init_function_from_entry_points()
    Configuration.load(application, configuration_groups=configuration_groups,
                       **kwargs)
    BlokManager.load()
    db_name = Configuration.get('db_name')
    db_template_name = Configuration.get('db_template_name', None)
    url = Configuration.get('get_url')(db_name=db_name)
    create_database(url, template=db_template_name)
    registry = RegistryManager.get(db_name)
    if registry is None:
        return

    if Configuration.get('install_all_bloks'):
        bloks = registry.System.Blok.list_by_state('uninstalled')
    else:
        install_bloks = Configuration.get('install_bloks') or []
        iou_bloks = Configuration.get('install_or_update_bloks') or []
        bloks = list(set(install_bloks + iou_bloks))

    registry.upgrade(install=bloks)
    registry.commit()
    registry.close()


def updatedb(application, configuration_groups, **kwargs):
    """ Update an existing database

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    format_configuration(configuration_groups, 'install-bloks',
                         'uninstall-bloks', 'update-bloks',
                         'install-or-update-bloks')

    registry = anyblok.start(application,
                             configuration_groups=configuration_groups,
                             loadwithoutmigration=True, **kwargs)

    installed_bloks = registry.System.Blok.list_by_state('installed')
    required_install_bloks = []
    required_update_bloks = []
    for blok in (Configuration.get('install_or_update_bloks') or []):
        if blok in installed_bloks:
                required_update_bloks.append(blok)
        else:
            required_install_bloks.append(blok)

    if Configuration.get('install_all_bloks'):
        install_bloks = registry.System.Blok.list_by_state('uninstalled')
    else:
        install_bloks = Configuration.get('install_bloks') or []
        install_bloks = list(set(install_bloks + required_install_bloks))

    if Configuration.get('update_all_bloks'):
        update_bloks = registry.System.Blok.list_by_state('installed')
    else:
        update_bloks = Configuration.get('update_bloks') or []
        update_bloks = list(set(update_bloks + required_update_bloks))

    uninstall_bloks = Configuration.get('uninstall_bloks')

    if registry:
        registry.update_blok_list()  # case, new blok added
        registry.upgrade(install=install_bloks, update=update_bloks,
                         uninstall=uninstall_bloks)
        registry.commit()
        registry.close()


def run_exit(application, configuration_groups, **kwargs):
    """Run nose unit test for the registry

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    format_configuration(configuration_groups, 'unittest')
    registry = anyblok.start(application,
                             configuration_groups=configuration_groups,
                             useseparator=True, unittest=True, **kwargs)

    defaultTest = []
    if registry:
        installed_bloks = registry.System.Blok.list_by_state("installed")
        selected_bloks = Configuration.get('selected_bloks')
        if not selected_bloks:
            selected_bloks = installed_bloks

        unwanted_bloks = Configuration.get('unwanted_bloks') or []

        defaultTest = [path
                       for blok in installed_bloks
                       if blok in selected_bloks and blok not in unwanted_bloks
                       for path in [join(BlokManager.getPath(blok), 'tests')]
                       if exists(path)]
        registry.close()  # free the registry to force create it again

    sys.exit(main(defaultTest=defaultTest))


def interpreter(application, configuration_groups, **kwargs):
    """Execute a script or open an interpreter

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    format_configuration(configuration_groups, 'interpreter')
    registry = anyblok.start(application,
                             configuration_groups=configuration_groups,
                             **kwargs)
    if registry:
        registry.commit()
        python_script = Configuration.get('python_script')
        if python_script:
            with open(python_script, "r") as fh:
                exec(fh.read(), None, locals())
        else:
            try:
                from IPython import embed
                embed()
            except ImportError:
                import code
                code.interact(local=locals())


def cron_worker(application, configuration_groups, **kwargs):
    """Execute a cron worker

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    registry = anyblok.start(application,
                             configuration_groups=configuration_groups,
                             **kwargs)
    if registry:
        registry.commit()
        registry.System.Cron.run()


def registry2doc(application, configuration_groups, **kwargs):
    """Return auto documentation for the registry

    :param application: name of the application
    :param configuration_groups: list configuration groupe to load
    :param \**kwargs: ArgumentParser named arguments
    """
    format_configuration(configuration_groups, 'doc', 'schema')
    registry = anyblok.start(application,
                             configuration_groups=configuration_groups,
                             **kwargs)
    if registry:
        registry.commit()
        doc = registry.Documentation()
        doc.auto_doc()
        if Configuration.get('doc_format') == 'RST':
            with open(Configuration.get('doc_output'), 'w') as fp:
                doc.toRST(fp)
        elif Configuration.get('doc_format') == 'UML':
            format_ = Configuration.get('schema_format')
            name_ = Configuration.get('schema_output')
            dot = ModelSchema(name_, format=format_)
            doc.toUML(dot)
            dot.save()
        elif Configuration.get('doc_format') == 'SQL':
            format_ = Configuration.get('schema_format')
            name_ = Configuration.get('schema_output')
            dot = SQLSchema(name_, format=format_)
            doc.toSQL(dot)
            dot.save()


def anyblok_createdb():
    createdb('createdb', ['logging'])


def anyblok_updatedb():
    updatedb('updatedb', ['logging'])


def anyblok_nose():
    run_exit("nose", ['logging'])


def anyblok_interpreter():
    interpreter('interpreter', ['logging'])


def anyblok_cron_worker():
    cron_worker('worker', ['logging'])


def anyblok2doc():
    registry2doc('autodoc', ['logging'])



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import BlokManager, Blok, BlokManagerException
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.registry import RegistryException, RegistryConflictingException


class TestBlokManager(TestCase):

    def tearDown(self):
        super(TestBlokManager, self).tearDown()
        BlokManager.unload()

    def test_load_anyblok(self):
        BlokManager.load()
        if not BlokManager.list():
            self.fail('No blok load')
        if not BlokManager.has('anyblok-core'):
            self.fail("The blok 'anyblok-core' is missing")

        BlokManager.get('anyblok-core')

    def test_load_with_invalid_blok_group(self):
        try:
            BlokManager.load(entry_points=('Invalid blok group',))
            self.fail('Load with invalid blok group')
        except BlokManagerException:
            pass

    def test_load_without_blok_group(self):
        try:
            BlokManager.load(entry_points=())
            self.fail('No watchdog to load without blok groups')
        except BlokManagerException:
            pass

    def test_reload(self):
        BlokManager.load()
        BlokManager.set('invalid blok', None)
        BlokManager.get('invalid blok')
        BlokManager.reload()
        try:
            BlokManager.get('invalid blok')
            self.fail("Reload classmethod doesn't reload the bloks")
        except BlokManagerException:
            pass

    def test_reload_without_load(self):
        try:
            BlokManager.reload()
            self.fail('No exception when reload without previously load')
        except BlokManagerException:
            pass

    def test_get_invalid_blok(self):
        try:
            BlokManager.load()
            BlokManager.get('invalid blok')
            self.fail('No exception when get invalid blok')
        except BlokManagerException:
            pass

    def test_set(self):
        blok_name = 'ABlok'
        BlokManager.set(blok_name, Blok)
        self.assertTrue(BlokManager.has(blok_name))

    def test_set_two_time(self):
        blok_name = 'ABlok'
        BlokManager.set(blok_name, Blok)
        try:
            BlokManager.set(blok_name, Blok)
            self.fail("No watch dog for set two time the same blok")
        except BlokManagerException:
            pass


class TestBlok(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_blok_exist(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok1')
        if not query.count():
            self.fail('No blok found')

        testblok1 = query.first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, None)
        self.assertEqual(testblok1.short_description, '')
        self.assertEqual(testblok1.long_description, '')

    def test_install(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')

    def test_install_an_installed_blok(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        try:
            registry.upgrade(install=('test-blok1',))
            self.fail('No watchdog to install an installed blok')
        except RegistryException:
            pass

    def test_uninstall(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, None)

    def test_uninstall_an_uninstalled_blok(self):
        registry = self.init_registry(None)
        try:
            registry.upgrade(uninstall=('test-blok1',))
            self.fail('No watchdog to uninstall an uninstalled blok')
        except RegistryException:
            pass

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok1.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '2.0.0')

    def test_update_an_uninstalled_blok(self):
        registry = self.init_registry(None)
        try:
            registry.upgrade(update=('test-blok1',))
            self.fail('No watchdog to update an uninstalled blok')
        except RegistryException:
            pass


class TestBlokRequired(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_blok_exist(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok2')
        if not query.count():
            self.fail('No blok found')

        testblok2 = query.first()
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        self.assertEqual(testblok2.short_description, 'Test blok2')

    def test_install_1by1(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')

    def test_install(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')

    def test_uninstall(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok2',))
        registry.upgrade(uninstall=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)

    def test_uninstall_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok2',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertIsNone(testblok1.installed_version)
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        registry.upgrade(update=('test-blok2',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.installed_version, '2.0.0')

    def test_update_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '2.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.installed_version, '2.0.0')


class TestBlokRequired2(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_blok_exist(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok3')
        if not query.count():
            self.fail('No blok found')

        testblok2 = query.first()
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        long_description = ".. This file is a part of the AnyBlok project\n.."
        long_description += "\n..    Copyright (C) 2014 Jean-Sebastien SUZANNE"
        long_description += " <jssuzanne@anybox.fr>\n..\n.. This Source Code "
        long_description += "Form is subject to the terms of the Mozilla "
        long_description += "Public License,\n.. v. 2.0. If a copy of the MPL "
        long_description += "was not distributed with this file,You can\n.. "
        long_description += "obtain one at http://mozilla.org/MPL/2.0/.\n\n"
        long_description += "Test blok3\n"
        self.assertEqual(testblok2.long_description, long_description)

    def test_install_1by1(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        self.assertEqual(testblok3.state, 'uninstalled')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertIsNone(testblok3.installed_version)
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')
        self.assertEqual(testblok3.state, 'uninstalled')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertIsNone(testblok3.installed_version)
        registry.upgrade(install=('test-blok3',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')
        self.assertEqual(testblok3.state, 'installed')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertEqual(testblok3.installed_version, '1.0.0')

    def test_install(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok3',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')
        self.assertEqual(testblok3.state, 'installed')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertEqual(testblok3.installed_version, '1.0.0')

    def test_uninstall(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertEqual(testblok2.installed_version, '1.0.0')
        self.assertEqual(testblok3.state, 'uninstalled')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertIsNone(testblok3.installed_version)

    def test_uninstall_first_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        self.assertEqual(testblok3.state, 'uninstalled')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertIsNone(testblok3.installed_version)

    def test_uninstall_all_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertIsNone(testblok1.installed_version)
        self.assertEqual(testblok2.state, 'uninstalled')
        self.assertEqual(testblok2.version, '1.0.0')
        self.assertIsNone(testblok2.installed_version)
        self.assertEqual(testblok3.state, 'uninstalled')
        self.assertEqual(testblok3.version, '1.0.0')
        self.assertIsNone(testblok3.installed_version)

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok3',))
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.installed_version, '1.0.0')
        self.assertEqual(testblok3.state, 'installed')
        self.assertEqual(testblok3.installed_version, '2.0.0')

    def test_update_first_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok2',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.installed_version, '2.0.0')
        self.assertEqual(testblok3.state, 'installed')
        self.assertEqual(testblok3.installed_version, '2.0.0')

    def test_update_all_required(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '2.0.0')
        self.assertEqual(testblok2.state, 'installed')
        self.assertEqual(testblok2.installed_version, '2.0.0')
        self.assertEqual(testblok3.state, 'installed')
        self.assertEqual(testblok3.installed_version, '2.0.0')


class TestBlokConditionnal(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_install_1by1(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok4.state, 'uninstalled')
        self.assertEqual(testblok4.version, '1.0.0')
        self.assertIsNone(testblok4.installed_version)
        self.assertEqual(testblok5.state, 'uninstalled')
        self.assertEqual(testblok5.version, '1.0.0')
        self.assertIsNone(testblok5.installed_version)
        registry.upgrade(install=('test-blok4',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok4.state, 'installed')
        self.assertEqual(testblok4.version, '1.0.0')
        self.assertEqual(testblok4.installed_version, '1.0.0')
        self.assertEqual(testblok5.state, 'installed')
        self.assertEqual(testblok5.version, '1.0.0')
        self.assertEqual(testblok5.installed_version, '1.0.0')

    def test_install(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok5',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok4.state, 'installed')
        self.assertEqual(testblok4.version, '1.0.0')
        self.assertEqual(testblok4.installed_version, '1.0.0')
        self.assertEqual(testblok5.state, 'installed')
        self.assertEqual(testblok5.version, '1.0.0')
        self.assertEqual(testblok5.installed_version, '1.0.0')

    def test_uninstall(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok5',))
        try:
            registry.upgrade(uninstall=('test-blok5',))
            self.fail('No watchdog to uninstall conditionnal blok')
        except RegistryException:
            pass

    def test_uninstall_conditionnal(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok5',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertIsNone(testblok1.installed_version)
        self.assertEqual(testblok4.state, 'installed')
        self.assertEqual(testblok4.version, '1.0.0')
        self.assertEqual(testblok4.installed_version, '1.0.0')
        self.assertEqual(testblok5.state, 'uninstalled')
        self.assertEqual(testblok5.version, '1.0.0')
        self.assertIsNone(testblok5.installed_version)

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok5',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        testblok1.version = '2.0.0'
        testblok4.version = '2.0.0'
        testblok5.version = '2.0.0'
        registry.upgrade(update=('test-blok5',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok4.state, 'installed')
        self.assertEqual(testblok4.installed_version, '1.0.0')
        self.assertEqual(testblok5.state, 'installed')
        self.assertEqual(testblok5.installed_version, '2.0.0')

    def test_update_conditionnal(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok5',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        testblok1.version = '2.0.0'
        testblok4.version = '2.0.0'
        testblok5.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '2.0.0')
        self.assertEqual(testblok4.state, 'installed')
        self.assertEqual(testblok4.installed_version, '1.0.0')
        self.assertEqual(testblok5.state, 'installed')
        self.assertEqual(testblok5.installed_version, '2.0.0')


class TestBlokOptional(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_install_1by1(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok6.state, 'uninstalled')
        self.assertEqual(testblok6.version, '1.0.0')
        self.assertIsNone(testblok6.installed_version)
        registry.upgrade(install=('test-blok6',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok6.state, 'installed')
        self.assertEqual(testblok6.version, '1.0.0')
        self.assertEqual(testblok6.installed_version, '1.0.0')

    def test_install(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok6',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok6.state, 'installed')
        self.assertEqual(testblok6.version, '1.0.0')
        self.assertEqual(testblok6.installed_version, '1.0.0')

    def test_uninstall(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok6',))
        registry.upgrade(uninstall=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok6.state, 'uninstalled')
        self.assertEqual(testblok6.version, '1.0.0')
        self.assertIsNone(testblok6.installed_version)

    def test_uninstall_optional(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok6',))
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok6.version = '2.0.0'
        registry.upgrade(uninstall=('test-blok1',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok1.version, '1.0.0')
        self.assertIsNone(testblok1.installed_version)
        self.assertEqual(testblok6.state, 'installed')
        self.assertEqual(testblok6.version, '2.0.0')
        self.assertEqual(testblok6.installed_version, '2.0.0')

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok1.version = '2.0.0'
        testblok6.version = '2.0.0'
        registry.upgrade(update=('test-blok6',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '1.0.0')
        self.assertEqual(testblok6.state, 'installed')
        self.assertEqual(testblok6.installed_version, '2.0.0')

    def test_update_optional(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok1.version = '2.0.0'
        testblok6.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok1.installed_version, '2.0.0')
        self.assertEqual(testblok6.state, 'installed')
        self.assertEqual(testblok6.installed_version, '2.0.0')


class TestBlokConflicting(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_marked_as_conflicting(self):
        blok1 = BlokManager.get('test-blok1')
        blok13 = BlokManager.get('test-blok13')
        self.assertEqual(blok13.conflicting, ['test-blok1'])
        self.assertEqual(blok1.conflicting_by, ['test-blok13'])

    def test_install_1by1(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok1',))
        with self.assertRaises(RegistryConflictingException):
            registry.upgrade(install=('test-blok13',))

    def test_install_both(self):
        registry = self.init_registry(None)
        with self.assertRaises(RegistryConflictingException):
            registry.upgrade(install=('test-blok1', 'test-blok13'))

    def test_uninstall_first_and_intall_another_in_two_step(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok13 = Blok.query().filter(Blok.name == 'test-blok13').first()
        registry.upgrade(install=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok13.state, 'uninstalled')
        registry.upgrade(uninstall=('test-blok1',))
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok13.state, 'uninstalled')
        registry.upgrade(install=('test-blok13',))
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok13.state, 'installed')

    def test_replace_blok_by_another_in_one_step(self):
        registry = self.init_registry(None)
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok13 = Blok.query().filter(Blok.name == 'test-blok13').first()
        registry.upgrade(install=('test-blok1',))
        self.assertEqual(testblok1.state, 'installed')
        self.assertEqual(testblok13.state, 'uninstalled')
        registry.upgrade(uninstall=('test-blok1',), install=('test-blok13',))
        self.assertEqual(testblok1.state, 'uninstalled')
        self.assertEqual(testblok13.state, 'installed')


class TestBlokOrder(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def check_order(self, registry, mode, wanted):
        Test = registry.Test
        bloks = Test.query().filter(Test.mode == mode).all().blok
        for _wanted in wanted:
            self.assertIn(_wanted, bloks)

    def test_install(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        self.check_order(registry, 'install', [
            'test-blok1', 'test-blok2', 'test-blok3'])

    def test_uninstall(self):
        from anyblok.blok import Blok, BlokManager
        old_uninstall = Blok.uninstall

        uninstalled = []

        def uninstall(self):
            cls = self.__class__
            uninstalled.extend(
                [x for x, y in BlokManager.bloks.items() if y is cls])

        try:
            Blok.uninstall = uninstall
            registry = self.init_registry(None)
            registry.upgrade(install=('test-blok3',))
            registry.upgrade(uninstall=('test-blok1',))
            self.assertEqual(uninstalled, [
                'test-blok3', 'test-blok2', 'test-blok1'])
        finally:
            Blok.uninstall = old_uninstall

    def test_update(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(update=('test-blok1',))
        self.check_order(registry, 'update', [
            'test-blok1', 'test-blok2', 'test-blok3'])

    def test_load(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok3',))
        self.check_order(registry, 'load', [
            'anyblok-core', 'test-blok1', 'test-blok2', 'test-blok3'])


class TestBlokModel(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_remove_foreign_key_after_uninstallation(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok7', 'test-blok8'))
        t2 = registry.Test2.insert(label="test2")
        registry.Test.insert(label="Test1", test2=t2.id)
        from sqlalchemy.exc import IntegrityError
        with self.assertRaises(IntegrityError):
            with registry.begin_nested():
                registry.Test2.query().delete()

        registry.upgrade(uninstall=('test-blok8',))
        registry.Test2.query().delete()


class TestBlokSession(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_session_with_no_change(self):
        registry = self.init_registry(None)
        Session = registry.Session
        registry.upgrade(install=('test-blok1',))
        self.assertIs(Session, registry.Session)

    def test_session_with_change_query(self):
        registry = self.init_registry(None)
        Session = registry.Session
        registry.upgrade(install=('test-blok11',))
        self.assertIsNot(Session, registry.Session)

    def test_session_with_change_session(self):
        registry = self.init_registry(None)
        Session = registry.Session
        registry.upgrade(install=('test-blok12',))
        self.assertIsNot(Session, registry.Session)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.schema import ForeignKey
from anyblok.common import anyblok_column_prefix


class ModelReprException(Exception):
    """Exception for Model attribute"""


class ModelAttributeException(Exception):
    """Exception for Model attribute"""


class ModelAttributeAdapterException(Exception):
    """Exception for Model attribute adapter"""


class MapperException(AttributeError):
    """ Simple Exception for Mapper """


class FakeColumn:
    db_column_name = None

    def update_description(self, registry, model, res):
        pass


class FakeRelationShip:

    def __init__(self, mapper):
        self.mapper = mapper

    def update_description(self, registry, model, res):
        pass


class ModelAttribute:
    """The Model attribute represente the using of a declared attribute, in the
    goal of get the real attribute after of the foreign_key::

        ma = ModelAttribute('registry name', 'attribute name')

    """

    def __repr__(self):
        return "%s => %s" % (self.model_name, self.attribute_name)

    def __str__(self):
        return "%s => %s" % (self.model_name, self.attribute_name)

    def __init__(self, model_name, attribute_name):
        if not model_name or not attribute_name:
            raise ModelAttributeException(
                "model_name (%s) and attribute_name (%s) are "
                "required" % (model_name, attribute_name))

        self.model_name = model_name
        self.attribute_name = attribute_name
        self._options = {}

    def get_attribute(self, registry, usehybrid=True):
        """Return the assembled attribute, the model need to be assembled

        :param registry: instance of the registry
        :param usehybrid: if True return the hybrid property if exist
        :rtype: instance of the attribute
        :exceptions: ModelAttributeException
        """
        if self.model_name not in registry.loaded_namespaces:
            raise ModelAttributeException(
                "Unknow model %r, maybe the model doesn't exist or is not"
                "assembled yet" % self.model_name)

        Model = registry.get(self.model_name)
        if not hasattr(Model, self.attribute_name):
            raise ModelAttributeException(
                "Model %r has not get %r attribute" % (
                    self.model_name, self.attribute_name))

        attribute_name = self.attribute_name
        if not usehybrid:
            if attribute_name in Model.hybrid_property_columns:
                attribute_name = anyblok_column_prefix + attribute_name

        return getattr(Model, attribute_name)

    def get_fk_column(self, registry):
        """Return the foreign key which represent the attribute in the data
        base

        :param registry: instance of the sqlalchemy ForeignKey
        :rtype: instance of the attribute
        """
        mapper = self.get_fk_mapper(registry)
        if mapper:
            return mapper.attribute_name

        return None

    def get_fk_mapper(self, registry):
        """Return the foreign key which represent the attribute in the data
        base

        :param registry: instance of the sqlalchemy ForeignKey
        :rtype: instance of the attribute
        """
        Model = self.check_model_in_first_step(registry)
        try:
            column_name = self.check_column_in_first_step(registry, Model)
            if Model[column_name].foreign_key:
                return Model[column_name].foreign_key
        except ModelAttributeException:
            pass

        return None

    def get_fk(self, registry):
        """Return the foreign key which represent the attribute in the data
        base

        :param registry: instance of the sqlalchemy ForeignKey
        :rtype: instance of the attribute
        """
        return ForeignKey(self.get_fk_name(registry), **self._options)

    def options(self, **kwargs):
        """Add foreign key options to create the sqlalchemy ForeignKey

        :param \**kwargs: options
        :rtype: the instance of the ModelAttribute
        """
        self._options.update(kwargs)
        return self

    def get_fk_name(self, registry):
        """Return the name of the foreign key

        the need of foreign key may be before the creation of the model in
        the registry, so we must use the first step of assembling

        :param registry: instance of the registry
        :rtype: str of the foreign key (tablename.columnname)
        :exceptions: ModelAttributeException
        """
        Model = self.check_model_in_first_step(registry)
        column_name = self.check_column_in_first_step(registry, Model)
        tablename = Model['__tablename__']
        if Model[self.attribute_name].db_column_name:
            column_name = Model[self.attribute_name].db_column_name

        return tablename + '.' + column_name

    def get_fk_remote(self, registry):
        Model = self.check_model_in_first_step(registry)
        column_name = self.check_column_in_first_step(registry, Model)
        remote = Model[column_name].foreign_key
        if not remote:
            return None

        return remote.get_fk_name(registry)

    def add_fake_column(self, registry):
        Model = self.check_model_in_first_step(registry)
        if self.attribute_name in registry.loaded_namespaces_first_step:
            return

        Model[self.attribute_name] = FakeColumn()

    def add_fake_relationship(self, registry, namespace, fieldname):
        Model = self.check_model_in_first_step(registry)
        if self.attribute_name in registry.loaded_namespaces_first_step:
            return

        Model[self.attribute_name] = FakeRelationShip(ModelAttribute(
            namespace, fieldname))

    def get_column_name(self, registry):
        """Return the name of the column

        the need of foreign key may be before the creation of the model in
        the registry, so we must use the first step of assembling

        :param registry: instance of the registry
        :rtype: str of the foreign key (tablename.columnname)
        :exceptions: ModelAttributeException
        """
        Model = self.check_model_in_first_step(registry)
        column_name = self.check_column_in_first_step(registry, Model)
        if hasattr(Model[self.attribute_name], 'db_column_name'):
            if Model[self.attribute_name].db_column_name:
                column_name = Model[self.attribute_name].db_column_name

        return column_name

    def check_model_in_first_step(self, registry):
        if self.model_name not in registry.loaded_namespaces_first_step:
            raise ModelAttributeException(
                "Unknow model %r" % self.model_name)

        Model = registry.loaded_namespaces_first_step[self.model_name]
        if len(Model.keys()) == 1:
            # No column found, so is not an sql model
            raise ModelAttributeException(
                "The Model %r is not an SQL Model" % self.model_name)
        return Model

    def check_column_in_first_step(self, registry, Model):
        if self.attribute_name not in Model:
            raise ModelAttributeException(
                "the Model %r has not got attribute %r" % (
                    self.model_name, self.attribute_name))

        return self.attribute_name

    def is_declared(self, registry):
        Model = self.check_model_in_first_step(registry)
        if self.attribute_name not in Model:
            return False

        return True

    def native_type(self, registry):
        Model = self.check_model_in_first_step(registry)
        column_name = self.check_column_in_first_step(registry, Model)
        return Model[column_name].native_type()


class ModelRepr:
    """Pseudo class to represent a model
    ::

        mr = ModelRepr('registry name')
    """
    def __init__(self, model_name):
        if not model_name:
            raise ModelReprException(
                "model_name (%s) is required" % model_name)

        self.model_name = model_name

    def __str__(self):
        return self.model_name

    def check_model(self, registry):
        """Check if the model exist else raise an exception

        :param registry: instance of the registry
        :rtype: dict which represent the first step of the model
        :exceptions: ModelReprException
        """
        if self.model_name not in registry.loaded_namespaces_first_step:
            raise ModelReprException("Model %r unexisting" % self.model_name)

        return registry.loaded_namespaces_first_step[self.model_name]

    def tablename(self, registry):
        """Return the  real tablename of the Model

        :param registry: instance of the registry
        :rtype: string
        """
        Model = self.check_model(registry)
        return Model['__tablename__']

    def primary_keys(self, registry):
        """Return the  of the primary keys

        :param registry: instance of the registry
        :rtype: list of ModelAttribute
        """
        from anyblok.column import Column
        Model = self.check_model(registry)
        pks = []
        for k, v in Model.items():
            if isinstance(v, Column):
                if 'primary_key' in v.kwargs:
                    pks.append(ModelAttribute(self.model_name, k))

        return pks

    def foreign_keys_for(self, registry, remote_model):
        """Return the  of the primary keys

        :param registry: instance of the registry
        :rtype: list of ModelAttribute
        """
        from anyblok.column import Column
        Model = self.check_model(registry)
        fks = []
        for k, v in Model.items():
            if isinstance(v, Column):
                if v.foreign_key:
                    if v.foreign_key.model_name == remote_model:
                        fks.append(ModelAttribute(self.model_name, k))

        return fks


def ModelAttributeAdapter(Model):
    """ Return a ModelAttribute

    :param Model: ModelAttribute or string ('registry name'=>'attribute name')
    :rtype: instance of ModelAttribute
    :exceptions: ModelAttributeAdapterException
    """
    if isinstance(Model, str):
        if '=>' not in Model:
            raise ModelAttributeAdapterException(
                "Wrong value %r impossible to find model and attribtue" % (
                    Model))
        model, attribute = Model.split('=>')
        return ModelAttribute(model, attribute)
    else:
        return Model


def ModelAdapter(Model):
    """ Return a ModelRepr

    :param Model: ModelRepr or string
    :rtype: instance of ModelRepr
    :exceptions: ModelAdapterException
    """
    if isinstance(Model, str):
        return ModelRepr(Model)
    elif isinstance(Model, ModelRepr):
        return Model
    else:
        return ModelRepr(Model.__registry_name__)


class ModelMapper:

    sqlalchemy_known_events = [
        'after_delete', 'after_insert', 'after_update', 'append_result',
        'before_delete', 'before_insert', 'before_update', 'create_instance',
        'expire', 'first_init', 'init', 'load', 'refresh',
    ]

    def __init__(self, mapper, event, *args, **kwargs):
        if isinstance(mapper, str):
            self.model = ModelRepr(mapper)
        elif isinstance(mapper, ModelRepr):
            self.model = mapper
        elif hasattr(mapper, '__registry_name__'):
            self.model = ModelRepr(mapper.__registry_name__)

        self.event = event
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def capable(cls, mapper):
        if isinstance(mapper, str):
            return True
        elif isinstance(mapper, ModelRepr):
            return True
        elif hasattr(mapper, '__registry_name__'):
            return True

        return False

    def listen(self, method):
        if self.event in self.sqlalchemy_known_events:
            method.is_an_sqlalchemy_event_listener = True
            method.sqlalchemy_listener = self
        else:
            method.is_an_event_listener = True
            method.model = self.model.model_name
            method.event = self.event

    def mapper(self, registry, namespace, **kwargs):
        model = self.model
        if self.model.model_name.upper() == 'SELF':
            model = ModelRepr(namespace)

        model.check_model(registry)
        return registry.get(model.model_name)


class ModelAttributeMapper:

    def __init__(self, mapper, event, *args, **kwargs):
        if isinstance(mapper, str):
            self.attribute = ModelAttribute(*mapper.split('=>'))
        elif isinstance(mapper, ModelAttribute):
            self.attribute = mapper

        self.event = event
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def capable(cls, mapper):
        if isinstance(mapper, str):
            if '=>' in mapper:
                model, column = mapper.split('=>')
                if model and column:
                    return True
        elif isinstance(mapper, ModelAttribute):
            return True

        return False

    def listen(self, method):
        method.is_an_sqlalchemy_event_listener = True
        method.sqlalchemy_listener = self

    def mapper(self, registry, namespace, usehybrid=True):
        attribute = self.attribute
        if self.attribute.model_name.upper() == 'SELF':
            attribute = ModelAttribute(
                namespace, self.attribute.attribute_name)

        return attribute.get_attribute(registry, usehybrid=usehybrid)


def MapperAdapter(mapper, *args, **kwargs):
    if ModelAttributeMapper.capable(mapper):
        return ModelAttributeMapper(mapper, *args, **kwargs)
    elif ModelMapper.capable(mapper):
        return ModelMapper(mapper, *args, **kwargs)
    else:
        raise MapperException("Unknow mapper type %r")



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations, classmethod_cache
from anyblok.field import FieldException
from anyblok.column import Column
from anyblok.mapper import FakeColumn, FakeRelationShip
from anyblok.relationship import RelationShip, Many2Many
from ..exceptions import SqlBaseException
from sqlalchemy.orm import aliased, ColumnProperty
from sqlalchemy.sql.expression import true
from sqlalchemy import or_, and_


class uniquedict(dict):

    def add_in_res(self, key, attrs):
        if key not in self:
            self[key] = []

        for attr in attrs:
            if attr not in self[key]:
                self[key].append(attr)


class SqlMixin:

    @classmethod
    def define_table_args(cls):
        return ()

    @classmethod
    def define_mapper_args(cls):
        return {}

    @classmethod
    def query(cls, *elements):
        """ Facility to do a SqlAlchemy query::

            query = MyModel.query()

        is equal at::

            query = self.registry.session.query(MyModel)

        :param elements: pass at the SqlAlchemy query, if the element is a
                         string then thet are see as field of the model
        :rtype: SqlAlchemy Query
        """
        res = []
        for f in elements:
            if isinstance(f, str):
                res.append(getattr(cls, f).label(f))
            else:
                res.append(f)

        if res:
            return cls.registry.query(*res)

        return cls.registry.query(cls)

    is_sql = True

    @classmethod
    def aliased(cls, *args, **kwargs):
        """ Facility to Apply an aliased on the model::

            MyModelAliased = MyModel.aliased()

        is equal at::

            from sqlalchemy.orm import aliased
            MyModelAliased = aliased(MyModel)

        :rtype: SqlAlchemy aliased of the model
        """
        return aliased(cls, *args, **kwargs)

    @classmethod
    def get_where_clause_from_primary_keys(cls, **pks):
        """ return the where clause to find object from pks

        :param \*\*pks: dict {primary_key: value, ...}
        :rtype: where clause
        :exception: SqlBaseException
        """
        _pks = cls.get_primary_keys()
        for pk in _pks:
            if pk not in pks:
                raise SqlBaseException("No primary key %s filled for %r" % (
                    pk, cls.__registry_name__))

        return [getattr(cls, k) == v for k, v in pks.items()]

    @classmethod
    def from_primary_keys(cls, **pks):
        """ return the instance of the model from the primary keys

        :param \*\*pks: dict {primary_key: value, ...}
        :rtype: instance of the model
        """
        where_clause = cls.get_where_clause_from_primary_keys(**pks)
        query = cls.query().filter(*where_clause)
        if query.count():
            return query.first()

        return None

    @classmethod
    def from_multi_primary_keys(cls, *pks):
        """ return the instances of the model from the primary keys

        :param \*pks: list of dict [{primary_key: value, ...}]
        :rtype: instances of the model
        """
        where_clause = []
        for _pks in pks:
            where_clause.append(cls.get_where_clause_from_primary_keys(**_pks))

        if not where_clause:
            return []

        where_clause = or_(*[and_(*x) for x in where_clause])

        query = cls.query().filter(where_clause)
        if query.count():
            return query.all()

        return []

    def to_primary_keys(self):
        """ return the primary keys and values for this instance

        :rtype: dict {primary key: value, ...}
        """
        pks = self.get_primary_keys()
        return {x: getattr(self, x) for x in pks}

    @classmethod
    def get_primary_keys(cls):
        """ return the name of the primary keys of the model

        :type: list of the primary keys name
        """
        C = cls.registry.System.Column
        model = cls.__registry_name__
        query = C.query()
        query = query.filter(C.model == model)
        query = query.filter(C.primary_key == true())
        return query.all().name

    @classmethod_cache()
    def _fields_description(cls):
        """ Return the information of the Field, Column, RelationShip """
        Field = cls.registry.System.Field
        res = {}
        for base in cls.__anyblok_bases__:
            query = Field.query().filter(Field.model == base.__registry_name__)
            res.update({x.name: x._description() for x in query.all()})

        query = Field.query().filter(Field.model == cls.__registry_name__)
        res.update({x.name: x._description() for x in query.all()})
        return res

    @classmethod
    def fields_description(cls, fields=None):
        res = cls._fields_description()
        if fields:
            return {x: y for x, y in res.items() if x in fields}

        return res

    @classmethod_cache()
    def get_hybrid_property_columns(cls):
        """Return the hybrid properties columns name from the Model and the
        inherited model if they come from polymorphisme
        """
        hybrid_property_columns = cls.hybrid_property_columns
        if 'polymorphic_identity' in cls.__mapper_args__:
            pks = cls.get_primary_keys()
            fd = cls.fields_description(*pks)
            for pk in pks:
                if fd[pk].get('model'):
                    Model = cls.registry.get(fd[pk]['model'])
                    hybrid_property_columns.extend(
                        Model.get_hybrid_property_columns())

        return hybrid_property_columns

    def _format_field(self, field):
        related_fields = None
        if isinstance(field, (tuple, list)):
            if len(field) == 1:
                related_fields = ()
            elif len(field) == 2:
                related_fields = field[1]
                if related_fields is None:
                    related_fields = ()
                elif not isinstance(related_fields, (tuple, list)):
                    raise SqlBaseException("%r the related fields wanted "
                                           "must be a tuple or empty or "
                                           "None value" % related_fields)
            else:
                raise SqlBaseException("%r the number of argument is "
                                       "wrong, waiting 1 or 2 arguments "
                                       "(name of the relation[, (related "
                                       "fields)])" % (field,))

            field = field[0]

        return field, related_fields

    def to_dict(self, *fields):
        """ Transform a record to the dict of value

        :param fields: list of fields to put in dict; if not selected, fields
            then take them all. A field is either one of these:

                * a string (which is the name of the field)
                * a 2-tuple if the field is a relationship (name of the field,
                  tuple of foreign model fields)

        :rtype: dict

        Here are some examples::

            =>> instance.to_dict()  # get all fields
            {"id": 1,
             "column1": "value 1",
             "column2": "value 2",
             "column3": "value 3",
             "relation1": {"relation_pk_1": 42, "relation_pk_2": "also 42"}
                                            # m2o or o2o : this is a dictionary
             "relation2": [{"id": 28}, {"id": 1}, {"id": 34}]
                                # o2m or m2m : this is a list of dictionaries
             }

            =>> instance.to_dict("column1", "column2", "relation1")
                        # get selected fields only (without any constraints)
            {"column1": "value 1",
             "column2": "value 2",
             "relation1": {"relation_pk_1": 42, "relation_pk_2": "also 42"}
             }

            =>> instance.to_dict("column1", "column2", (
                        # select fields to use in the relation related model
                "relation1", ("relation_pk1", "name", "value")
                            # there is no constraints in the choice of fields
                ))
            {"column1": "value",
             "column2": "value",
             "relation1": {"relation_pk_1": 42, "name": "H2G2", "value": "42"}
             }

            =>> instance.to_dict("column1", "column2", ("relation1", ))
            # or
            =>> instance.to_dict("column1", "column2", ("relation1", None))
            # or
            =>> instance.to_dict("column1", "column2", ("relation1", ()))
                                  # select all the fields of the relation ship
            {"column1": "value",
             "column2": "value",
             "relation1": {"relation_pk_1": 42, "name": "H2G2", "value": "42"}
             }

            =>> instance.to_dict("column1", "column2", (
                                        # select relation fields recursively
                "relation1", ("name", "value", (
                    "relation", ("a", "b", "c")
                    ))
                ))
            {"column1": "value",
             "column2": "value",
             "relation1": {"name": "H2G2", "value": "42", "relation": [
                 {"a": 10, "b": 20, "c": 30},
                 {"a": 11, "b": 22, "c": 33},
                 ]}
             }
        """
        result = {}
        cls = self.__class__
        fields = fields if fields else cls.fields_description().keys()

        for field in fields:
            # if field is ("relation_name", ("list", "of", "relation",
            # "fields")), deal with it.
            field, related_fields = self._format_field(field)
            # Get the actual data
            field_value, field_property = getattr(self, field), None
            try:
                field_property = getattr(getattr(cls, field), 'property', None)
            except FieldException:
                pass

            # Deal with this data
            if field_property is None:
                # it is the case of field function (hyprid property)
                result[field] = field_value
            elif field_value is None or type(field_property) == ColumnProperty:
                # If value is None, then do not go any further whatever
                # the column property tells you.
                result[field] = field_value
            else:
                # it is should be RelationshipProperty
                if related_fields is None:
                    # If there is no field list to the relation,
                    # use only primary keys
                    related_fields = field_property.mapper.entity
                    related_fields = related_fields.get_primary_keys()
                # One2One, One2Many, Many2One or Many2Many ?
                if field_property.uselist:
                    result[field] = [r.to_dict(*related_fields)
                                     for r in field_value]
                else:
                    result[field] = field_value.to_dict(*related_fields)

        return result

    @classmethod
    def getFieldType(cls, name):
        """Return the type of the column

        ::

            TheModel.getFieldType(nameOfTheColumn)

        this method take care if it is a polymorphic model or not

        :param name: name of the column
        :rtype: String, the name of the Type of column used
        """
        models = [x.__registry_name__ for x in cls.__anyblok_bases__]
        models.insert(0, cls.__registry_name__)
        Field = cls.registry.System.Field
        query = Field.query()
        query = query.filter(Field.name == name)
        query = query.filter(Field.model.in_(models))
        query = query.limit(1)
        return query.one().ftype


def get_model_information(registry, registry_name):
    model = registry.loaded_namespaces_first_step[registry_name]
    for depend in model['__depends__']:
        if depend != registry_name:
            for x, y in get_model_information(registry, depend).items():
                if x not in model:
                    model[x] = y

    return model


@Declarations.register(Declarations.Core)
class SqlBase(SqlMixin):
    """ this class is inherited by all the SQL model
    """

    def update(self, **values):
        """ Hight livel method to update the session for the instance
        ::

            self.update(val1=.., val2= ...)

        ..warning::

            the columns and values is passed as named arguments to show
            a difference with Query.update meth

        """
        for x, v in values.items():
            setattr(self, x, v)

        return 1 if values else 0

    @classmethod_cache()
    def find_remote_attribute_to_expire(cls, *fields):
        res = uniquedict()
        _fields = []
        _fields.extend(fields)
        model = get_model_information(cls.registry, cls.__registry_name__)
        while _fields:
            field = _fields.pop()
            field = field if isinstance(field, str) else field.name
            _field = model[field]

            if isinstance(_field, (Column, FakeColumn)):
                _fields.extend(x for x, y in model.items()
                               if (isinstance(y, RelationShip) and
                                   not isinstance(y, Many2Many))
                               for mapper in y.column_names
                               if mapper.attribute_name == field)
                if isinstance(_field, Column) and _field.foreign_key:
                    rmodel = cls.registry.loaded_namespaces_first_step[
                        _field.foreign_key.model_name]
                    for rc in [x for x, y in rmodel.items()
                               if isinstance(y, RelationShip)
                               for mapper in y.remote_columns
                               if mapper.attribute_name == field]:
                        rfield = rmodel[rc]
                        if isinstance(rfield, FakeRelationShip):
                            res.add_in_res(rfield.mapper.attribute_name, [rc])
                        elif (isinstance(rfield, RelationShip) and
                              'backref' in rfield.kwargs):
                            res.add_in_res(
                                rfield.kwargs['backref'][0], [rc])

            elif (isinstance(_field, RelationShip) and
                  not isinstance(_field, Many2Many) and
                  'backref' in _field.kwargs):
                res.add_in_res(field, [_field.kwargs['backref'][0]])
            elif isinstance(_field, FakeRelationShip):
                res.add_in_res(field, [_field.mapper.attribute_name])

        return res

    @classmethod_cache()
    def find_relationship(cls, *fields):
        """ Find column and relation ship link with the column or relationship
        passed in fields.

        :param \*fields: lists of the attribute name
        :rtype: list of the attribute name of the attribute and relation ship
        """
        res = []
        _fields = []
        _fields.extend(fields)
        model = get_model_information(cls.registry, cls.__registry_name__)
        while _fields:
            field = _fields.pop()
            if not isinstance(field, str):
                field = field.name

            if field in res:
                continue

            _field = model[field]
            res.append(field)
            if isinstance(_field, (Column, FakeColumn)):
                _fields.extend(x
                               for x, y in model.items()
                               if (isinstance(y, RelationShip) and
                                   not isinstance(y, Many2Many))
                               for mapper in y.column_names
                               if mapper.attribute_name == field)
            elif (isinstance(_field, RelationShip) and
                  not isinstance(_field, Many2Many)):
                for mapper in _field.column_names:
                    _fields.append(mapper.attribute_name)

        return res

    def expire_relationship_mapped(self, mappers):
        """ Expire the objects linked with this object, in function of
        the mappers definition
        """
        for field_name, rfields in mappers.items():
            fields = getattr(self, field_name)
            if not isinstance(fields, list):
                fields = [fields]

            for field in fields:
                if field is not None:
                    field.expire(*rfields)

    def refresh(self, *fields):
        """ Expire and reload all the attribute of the instance

        See: http://docs.sqlalchemy.org/en/latest/orm/session_api.html
        #sqlalchemy.orm.session.Session.refresh
        """
        self.registry.refresh(self, fields)

    def expunge(self):
        """Expunge the instance in the session"""
        self.registry.session.expunge(self)

    def expire(self, *fields):
        """ Expire the attribute of the instance, theses attributes will be
        load at the next  call of the instance

        see: http://docs.sqlalchemy.org/en/latest/orm/session_api.html
        #sqlalchemy.orm.session.Session.expire
        """
        self.registry.expire(self, fields)

    def delete(self, byquery=False, flush=True):
        """ Call the SqlAlchemy Query.delete method on the instance of the
        model::

            self.delete()

        is equal at::

            flush the session
            remove the instance of the session
            and expire all the session, to reload the relation ship

        """
        if byquery:
            cls = self.__class__
            cls.query().filter(*cls.get_where_clause_from_primary_keys(
                **self.to_primary_keys())).delete()
            self.expunge()
        else:
            model = self.registry.loaded_namespaces_first_step[
                self.__registry_name__]
            fields = model.keys()
            mappers = self.__class__.find_remote_attribute_to_expire(*fields)
            self.expire_relationship_mapped(mappers)
            self.registry.session.delete(self)

        if flush:
            self.registry.flush()

    @classmethod
    def insert(cls, **kwargs):
        """ Insert in the table of the model::

            MyModel.insert(...)

        is equal at::

            mymodel = MyModel(...)
            MyModel.registry.session.add(mymodel)
            MyModel.registry.flush()

        """
        instance = cls(**kwargs)
        cls.registry.add(instance)
        cls.registry.flush()
        return instance

    @classmethod
    def multi_insert(cls, *args):
        """ Insert in the table one or more entry of the model::

            MyModel.multi_insert([{...}, ...])

        the flush will be done only one time at the end of the insert

        :exception: SqlBaseException
        """
        instances = cls.registry.InstrumentedList()
        for kwargs in args:
            if not isinstance(kwargs, dict):
                raise SqlBaseException("multi_insert method wait list of dict")

            instance = cls(**kwargs)
            cls.registry.add(instance)
            instances.append(instance)

        if instances:
            cls.registry.flush()

        return instances

    @classmethod
    def precommit_hook(cls, method, *args, **kwargs):
        """ Same in the registry a hook to call just before the commit

        .. warning:: Only one instance of the hook is called before the commit

        :param method: the method to call on this model
        :param put_at_the_end_if_exist: If ``True`` the hook is move at the end
        """
        cls.registry.precommit_hook(
            cls.__registry_name__, method, *args, **kwargs)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Base classes for unit/integration tests with anyblok.

This module provides :class:`BlokTestCase`, which is the main one meant for
blok tests, and :class:`DBTestCase`, whose primary purpose is to test anyblok
itself, in so-called "framework tests".
"""

import unittest
import os
from anyblok.config import Configuration
from anyblok.registry import RegistryManager
from anyblok.blok import BlokManager
from anyblok.environment import EnvironmentManager
import sqlalchemy
from sqlalchemy.orm import clear_mappers
from sqlalchemy import event
from sqlalchemy_utils.functions import database_exists, create_database, orm
from copy import copy
from testfixtures import LogCapture as LC
from contextlib import contextmanager
from logging import getLogger, DEBUG, INFO, WARNING, ERROR, CRITICAL

logger = getLogger(__name__)


def drop_database(url):
    url = copy(sqlalchemy.engine.url.make_url(url))
    database = url.database
    if url.drivername.startswith('postgresql'):
        url.database = 'postgres'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    engine = sqlalchemy.create_engine(url)
    if engine.dialect.name == 'sqlite' and url.database != ':memory:':
        os.remove(url.database)
    else:
        text = 'DROP DATABASE {0}'.format(orm.quote(engine, database))
        cnx = engine.connect()
        cnx.execute("ROLLBACK")
        cnx.execute(text)
        cnx.execute("commit")
        cnx.close()


class TestCase(unittest.TestCase):
    """Common helpers, not meant to be used directly."""

    _transaction_case_teared_down = False

    @classmethod
    def init_configuration_manager(cls, **env):
        """ Initialise the configuration manager with environ variable
        to launch the test

        .. warning::

            For the moment we not use the environ variable juste constante

        :param prefix: prefix the database name
        :param env: add another dict to merge with environ variable
        """
        db_name = Configuration.get('db_name') or 'test_anyblok'
        db_driver_name = Configuration.get('db_driver_name') or 'postgresql'
        env.update({
            'db_name': db_name,
            'db_driver_name': db_driver_name,
        })
        Configuration.update(**env)

    @classmethod
    def createdb(cls, keep_existing=False):
        """Create the database specified in configuration.

        ::

            cls.init_configuration_manager()
            cls.createdb()

        :param keep_existing: If false drop the previous db before create it
        """
        url = Configuration.get('get_url')()
        db_template_name = Configuration.get('db_template_name', None)
        if database_exists(url):
            if keep_existing:
                return False

            drop_database(url)

        create_database(url, template=db_template_name)
        return True

    @classmethod
    def dropdb(cls):
        """Drop the database specified in configuration.

        ::

            cls.init_configuration_manager()
            cls.dropdb()

        """
        url = Configuration.get('get_url')()
        if database_exists(url):
            drop_database(url)

    @classmethod
    def additional_setting(cls):
        return dict(unittest=True)

    @classmethod
    def getRegistry(cls):
        """Return the registry for the test database.

        This assumes the database is created, and the registry has already
        been initialized::

            registry = self.getRegistry()

        :rtype: registry instance
        """
        additional_setting = cls.additional_setting()
        return RegistryManager.get(Configuration.get('db_name'),
                                   **additional_setting)

    def setUp(self):
        super(TestCase, self).setUp()
        self.addCleanup(self.callCleanUp)

    def callCleanUp(self):
        if not self._transaction_case_teared_down:
            self.cleanUp()

    def cleanUp(self):
        self.tearDown()

    def tearDown(self):
        """ Roll back the session """
        super(TestCase, self).tearDown()
        self._transaction_case_teared_down = True

    @contextmanager
    def Configuration(**values):
        """Add Configuration value only in the contextmanager
        ::

            with TestCase.Configuration(db_name='a db name'):
                self.assertEqual(Configuration.get('db_name'), 'a db name')

        :param \**values: values to update
        """
        try:
            old_configuration = Configuration.configuration.copy()
            Configuration.update(**values)
            yield
        finally:
            Configuration.configuration = old_configuration


class DBTestCase(TestCase):
    """Base class for tests that need to work on an empty database.

    .. warning::

        The database is created and dropped with each unit test

    For instance, this is the one used for Field, Column, RelationShip, and
    more generally core framework tests.

    The drawback of using this base class is that tests will be slow. The
    advantage is ultimate test isolation.

    Sample usage::

        from anyblok.tests.testcase import DBTestCase


        def simple_column(ColumnType=None, **kwargs):

            @Declarations.register(Declarations.Model)
            class Test:

                id = Declarations.Column.Integer(primary_key=True)
                col = ColumnType(**kwargs)


        class TestColumns(DBTestCase):

            def test_integer(self):
                Integer = Declarations.Column.Integer

                registry = self.init_registry(simple_column,
                                              ColumnType=Integer)
                test = registry.Test.insert(col=1)
                self.assertEqual(test.col, 1)

    """

    blok_entry_points = ('bloks',)
    """setuptools entry points to load blok """

    @classmethod
    def setUpClass(cls):
        """ Intialialise the configuration manager """

        super(DBTestCase, cls).setUpClass()
        cls.init_configuration_manager()
        if cls.createdb(keep_existing=True):
            BlokManager.load(entry_points=('bloks', 'test_bloks'))
            registry = cls.getRegistry()
            registry.commit()
            registry.close()
            BlokManager.unload()

    def setUp(self):
        """ Create a database and load the blok manager """
        self.registry = None
        super(DBTestCase, self).setUp()
        BlokManager.load(entry_points=self.blok_entry_points)

    def tearDown(self):
        """ Clear the registry, unload the blok manager and  drop the database
        """
        if self.registry:
            self.registry.close()

        RegistryManager.clear()
        BlokManager.unload()
        clear_mappers()
        super(DBTestCase, self).tearDown()

    def init_registry(self, function, **kwargs):
        """ call a function to filled the blok manager with new model

        :param function: function to call
        :param kwargs: kwargs for the function
        :rtype: registry instance
        """
        from copy import deepcopy
        loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
        if function is not None:
            EnvironmentManager.set('current_blok', 'anyblok-core')
            try:
                function(**kwargs)
            finally:
                EnvironmentManager.set('current_blok', None)

        try:
            self.registry = registry = self.__class__.getRegistry()
        finally:
            RegistryManager.loaded_bloks = loaded_bloks

        return registry


class BlokTestCase(unittest.TestCase):
    """Base class for tests meant to run on a preinstalled database.

    The tests written with this class don't need to start afresh on a new
    database, and therefore run much faster than those inheriting
    :class:`DBTestCase`. Instead, they expect the tested bloks to be already
    installed and up to date.

    The session gets rollbacked after each test.

    Such tests are appropriate for a typical blok developper workflow:

    * create and install the bloks once
    * run the tests of the blok under development repeatedly
    * upgrade the bloks in database when needed (schema change, update of
      dependencies)

    They are also appropriate for on the fly testing while installing the
    bloks: the tests of each blok get run on the database state they expect,
    before dependent (downstream) bloks, that could. e.g., alter the database
    schema, get themselves installed.
    This is useful to test a whole stack at once using only one
    database (typically in CI bots).

    Sample usage::

        from anyblok.tests.testcase import BlokTestCase


        class MyBlokTest(BlokTestCase):

            def test_1(self):
                # access to the registry by ``self.registry``
                ...

    """

    _transaction_case_teared_down = False
    registry = None
    """The instance of :class:`anyblok.registry.Registry`` to use in tests.

    The ``session_commit()`` method is disabled to avoid side effects from one
    test to the other.
    """

    @classmethod
    def additional_setting(cls):
        return dict(unittest=True)

    @classmethod
    def setUpClass(cls):
        """ Initialize the registry.
        """
        super(BlokTestCase, cls).setUpClass()
        additional_setting = cls.additional_setting()
        if cls.registry is None:
            cls.registry = RegistryManager.get(Configuration.get('db_name'),
                                               **additional_setting)

        cls.registry.commit()

    def setUp(self):
        super(BlokTestCase, self).setUp()
        self.addCleanup(self.callCleanUp)
        self.registry.begin_nested()  # add SAVEPOINT

        @event.listens_for(self.registry.session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            if transaction.nested and not transaction._parent.nested:
                session.expire_all()
                session.begin_nested()

    def callCleanUp(self):
        if not self._transaction_case_teared_down:
            self.cleanUp()

    def cleanUp(self):
        self.tearDown()

    def tearDown(self):
        """ Roll back the session """
        super(BlokTestCase, self).tearDown()
        try:
            self.registry.System.Cache.invalidate_all()
        except sqlalchemy.exc.InvalidRequestError:
            pass
        finally:
            self.registry.rollback()
            self.registry.session.close()

        self._transaction_case_teared_down = True


class LogCapture(LC):
    """Overwrite ``testfixtures.LogCapture`` to add some helper methods"""

    def get_messages(self, *levels):
        """Return the captured messages
        ::

            with LogCapture() as logs:
                messages = logs.get_messages(INFO, WARNING)

        :param \*levels: list of logging.level
        :rtype: list of formated message
        """
        return [
            self.format(r)
            for r in self.records
            if (not levels or r.levelno in levels)
        ]

    def get_debug_messages(self):
        """Return only the logging.DEBUG messages"""
        return self.get_messages(DEBUG)

    def get_info_messages(self):
        """Return only the logging.INFO messages"""
        return self.get_messages(INFO)

    def get_warning_messages(self):
        """Return only the logging.WARNING messages"""
        return self.get_messages(WARNING)

    def get_error_messages(self):
        """Return only the logging.ERROR messages"""
        return self.get_messages(ERROR)

    def get_critical_messages(self):
        """Return only the logging.CRITICAL messages"""
        return self.get_messages(CRITICAL)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import RegistryManager
from anyblok.environment import EnvironmentManager


class TestRegistryCore(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestRegistryCore, cls).setUpClass()
        RegistryManager.declare_core('test')
        RegistryManager.init_blok('testCore')
        EnvironmentManager.set('current_blok', 'testCore')

    @classmethod
    def tearDownClass(cls):
        super(TestRegistryCore, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testCore']
        RegistryManager.undeclare_core('test')

    def tearDown(self):
        super(TestRegistryCore, self).tearDown()
        RegistryManager.loaded_bloks['testCore']['Core']['test'] = []

    def assertInCore(self, *args):
        blokname = 'testCore'
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Core']['test']), len(args))
        for cls_ in args:
            hasCls = cls_ in blok['Core']['test']
            self.assertEqual(hasCls, True)

    def test_add_core(self):
        class test:
            pass

        RegistryManager.add_core_in_register('test', test)
        self.assertInCore(test)
        self.assertEqual(
            RegistryManager.has_core_in_register('testCore', 'test'),
            True)

    def test_remove_core(self):
        class test:
            pass

        def has_test_in_removed():
            core = RegistryManager.loaded_bloks['testCore']['removed']
            if test in core:
                return True

            return False

        RegistryManager.add_core_in_register('test', test)
        self.assertInCore(test)
        self.assertEqual(has_test_in_removed(), False)
        RegistryManager.remove_in_register(test)
        self.assertEqual(has_test_in_removed(), True)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.field import FieldException
from inspect import getmembers, ismethod, isfunction, ismodule, isclass


@Declarations.register(Declarations.Model.Documentation.Model)
class Attribute:

    def __init__(self, attribute):
        self.name, self.attribute = attribute

    def exist(self, model):
        return model.exist()

    @classmethod
    def filterAttribute(cls, model, name):
        if name in ('insert', 'update', 'to_primary_keys',
                    'to_dict', 'sqlalchemy_query_update',
                    'sqlalchemy_query_delete', 'query',
                    'precommit_hook', 'multi_insert', 'initialize_model',
                    'has_perm', 'has_model_perm',
                    'get_where_clause_from_primary_keys', 'get_primary_keys',
                    'get_model', 'from_primary_keys',
                    'from_multi_primary_keys', 'fire', 'fields_description',
                    '_fields_description', 'delete', 'aliased', '__init__',
                    'loaded_columns', 'loaded_fields', 'registry',
                    '_sa_class_manager', '_decl_class_registry'):
            return True

        return False

    @classmethod
    def getelements(cls, model):
        res = []
        Model = cls.registry.get(model.model.name)
        try:
            for k, v in getmembers(Model):
                if ismodule(v) or isclass(v):
                    continue

                if k.startswith('__'):
                    continue

                if cls.filterAttribute(model, k):
                    continue

                res.append((k, v))
        except FieldException:
            pass

        return res

    @classmethod
    def header2RST(cls, doc):
        title = "Attributes, methods and class methods"
        doc.write("%s\n%s\n\n" % (title, '~' * len(title)))

    @classmethod
    def footer2RST(cls, doc):
        pass

    def toRST(self, doc):
        doc.write('* ' + self.name + '\n\n')
        self.toRST_docstring(doc)

    def toRST_docstring(self, doc):
        if hasattr(self.attribute, '__doc__') and self.attribute.__doc__:
            doc.write(self.attribute.__doc__ + '\n\n')

    def toUML(self, dot, modelname):
        model = dot.get_class(modelname)
        if ismethod(self.attribute) or isfunction(self.attribute):
            model.add_method(self.name)
        else:
            model.add_property(self.name)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from datetime import datetime, timedelta
from ..exceptions import CronWorkerException


class TestCron(BlokTestCase):

    def test_add_worker_for(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        worker = Cron.add_worker_for(job)
        self.assertIs(job, worker.job)
        worker.join()

    def test_close_worker_with_success(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        worker = Cron.Worker(job)
        self.assertIsNone(job.done_at)
        Cron.close_worker_with_success(worker)
        self.assertIsNotNone(job.done_at)

    def test_close_worker_on_error(self):
        error = 'One error with this traceback'
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        worker = Cron.Worker(job)
        self.assertIsNone(job.done_at)
        self.assertIsNone(job.error)
        Cron.close_worker_on_error(worker, error)
        self.assertEqual(job.error, error)
        self.assertIsNone(job.done_at)

    def test_lock_one_record_no_existing_record(self):
        Cron = self.registry.System.Cron
        locked_job = Cron.lock_one_job()
        self.assertIsNone(locked_job)

    def test_lock_one_record_lock_existing_record(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        locked_job = Cron.lock_one_job()
        self.assertIs(locked_job, job)

    def test_lock_one_record_ended_existing_record(self):
        Cron = self.registry.System.Cron
        Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            done_at=datetime.now())
        locked_job = Cron.lock_one_job()
        self.assertIsNone(locked_job)

    def test_lock_one_record_upper_dated_existing_record(self):
        Cron = self.registry.System.Cron
        Cron.Job.insert(model="Model.System.Blok", method="list_by_state",
                        available_at=datetime.now() + timedelta(days=1))
        locked_job = Cron.lock_one_job()
        self.assertIsNone(locked_job)

    def test_lock_one_record_lock_existing_record2(self):
        Cron = self.registry.System.Cron
        Cron.Job.insert(model="Model.System.Blok", method="list_by_state",
                        done_at=datetime.now())
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        locked_job = Cron.lock_one_job()
        self.assertIs(locked_job, job)


class TestWorker(BlokTestCase):

    def test_get_model_record(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'primary_keys': {'name': 'anyblok-core'}})
        worker = Cron.Worker(job)
        record = worker.get_model_record()
        self.assertEqual(record.name, 'anyblok-core')

    def test_get_model_record_without_pks(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        worker = Cron.Worker(job)
        with self.assertRaises(CronWorkerException):
            worker.get_model_record()

    def test_get_model_record_with_invalid_pks(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'primary_keys': {'name': 'wrongblok'}})
        worker = Cron.Worker(job)
        with self.assertRaises(CronWorkerException):
            worker.get_model_record()

    def test_get_arg(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'args': 'name'})
        worker = Cron.Worker(job)
        args, kwargs = worker.get_args_and_kwargs()
        self.assertEqual(args, ('name',))
        self.assertEqual(kwargs, {})

    def test_get_args(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'args': ['name']})
        worker = Cron.Worker(job)
        args, kwargs = worker.get_args_and_kwargs()
        self.assertEqual(args, ('name',))
        self.assertEqual(kwargs, {})

    def test_get_invalid_kwargs(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'kwargs': 'name'})
        worker = Cron.Worker(job)
        with self.assertRaises(CronWorkerException):
            worker.get_args_and_kwargs()

    def test_call_method(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state")
        worker = Cron.Worker(job)
        res = worker.call_method()
        self.assertIsNone(res, None)

    def test_call_method2(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="list_by_state",
            params={'args': 'installed'})
        worker = Cron.Worker(job)
        res = worker.call_method()
        self.assertIsNotNone(res)

    def test_call_method3(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="get_short_description",
            is_a_class_method=False,
            params={'primary_keys': {'name': 'anyblok-core'}})
        worker = Cron.Worker(job)
        res = worker.call_method()
        ac = self.registry.System.Blok.from_primary_keys(name='anyblok-core')
        self.assertEqual(res, ac.short_description)

    def test_call_inexisting_method(self):
        Cron = self.registry.System.Cron
        job = Cron.Job.insert(
            model="Model.System.Blok", method="inexisting_method")
        worker = Cron.Worker(job)
        with self.assertRaises(CronWorkerException):
            worker.call_method()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import RegistryManager
from anyblok import Declarations
from anyblok.environment import EnvironmentManager
register = Declarations.register
unregister = Declarations.unregister
Mixin = Declarations.Mixin


class OneInterface:
    pass


class TestCoreInterfaceMixin(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestCoreInterfaceMixin, cls).setUpClass()
        RegistryManager.init_blok('testMixin')
        EnvironmentManager.set('current_blok', 'testMixin')

    @classmethod
    def tearDownClass(cls):
        super(TestCoreInterfaceMixin, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testMixin']

    def setUp(self):
        super(TestCoreInterfaceMixin, self).setUp()
        blokname = 'testMixin'
        RegistryManager.loaded_bloks[blokname]['Mixin'] = {
            'registry_names': []}

    def assertInMixin(self, *args):
        blokname = 'testMixin'
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Mixin']['Mixin.MyMixin']['bases']),
                         len(args))
        for cls_ in args:
            has = cls_ in blok['Mixin']['Mixin.MyMixin']['bases']
            self.assertEqual(has, True)

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testMixin']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Mixin, cls_=OneInterface, name_='MyMixin')
        self.assertEqual('Mixin', Mixin.MyMixin.__declaration_type__)
        self.assertInMixin(OneInterface)
        dir(Declarations.Mixin.MyMixin)

    def test_add_interface_with_decorator(self):

        @register(Mixin)
        class MyMixin:
            pass

        self.assertEqual('Mixin', Mixin.MyMixin.__declaration_type__)
        self.assertInMixin(MyMixin)

    def test_add_two_interface(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")
        self.assertInMixin(OneInterface)
        unregister(Mixin.MyMixin, OneInterface)

        blokname = 'testMixin'
        self.assertEqual(hasattr(Mixin, blokname), False)
        self.assertInMixin(OneInterface)
        self.assertInRemoved(OneInterface)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)
        unregister(Mixin.MyMixin, OneInterface)
        self.assertInMixin(MyMixin, OneInterface)
        self.assertInRemoved(OneInterface)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.config import Configuration
from anyblok import load_init_function_from_entry_points


load_init_function_from_entry_points(unittest=True)
Configuration.load_config_for_test()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.authorization.rule.modelaccess import ModelAccessRule


# Normally, this blok should be part of the edge
Declarations.AuthorizationBinding(Declarations.Model.Test2, ModelAccessRule())
